"""Sync I/O protocol implementation for the LM Studio remote access API."""

import itertools
import time
import weakref

from abc import abstractmethod
from concurrent.futures import Future as SyncFuture, ThreadPoolExecutor, as_completed
from contextlib import (
    contextmanager,
    ExitStack,
)
from types import EllipsisType, TracebackType
from typing import (
    Any,
    ContextManager,
    Generator,
    Iterable,
    Iterator,
    Callable,
    Generic,
    Sequence,
    Type,
    TypeAlias,
    TypeVar,
)
from typing_extensions import (
    # Native in 3.11+
    Self,
    # Native in 3.13+
    TypeIs,
)

import httpx

# Synchronous API still uses an async websocket (just in a background thread)
from httpx_ws import AsyncWebSocketSession

from .sdk_api import (
    LMStudioRuntimeError,
    LMStudioValueError,
    sdk_callback_invocation,
    sdk_public_api,
)
from .schemas import AnyLMStudioStruct, DictObject
from .history import (
    AssistantResponse,
    Chat,
    ChatHistoryDataDict,
    FileHandle,
    LocalFileInput,
    _LocalFileData,
    ToolCallRequest,
    ToolResultMessage,
)
from .json_api import (
    ActResult,
    AnyLoadConfig,
    AnyModelSpecifier,
    AvailableModelBase,
    ChannelEndpoint,
    ChannelHandler,
    ChatResponseEndpoint,
    ClientBase,
    ClientSession,
    CompletionEndpoint,
    DEFAULT_TTL,
    DownloadedModelBase,
    DownloadFinalizedCallback,
    DownloadProgressCallback,
    EmbeddingLoadModelConfig,
    EmbeddingLoadModelConfigDict,
    EmbeddingModelInfo,
    GetOrLoadEndpoint,
    LlmInfo,
    LlmLoadModelConfig,
    LlmLoadModelConfigDict,
    LlmPredictionConfig,
    LlmPredictionConfigDict,
    LlmPredictionFragment,
    LMStudioCancelledError,
    LMStudioClientError,
    LMStudioPredictionError,
    LMStudioTimeoutError,
    LMStudioWebsocket,
    LMStudioWebsocketError,
    LoadModelEndpoint,
    ModelDownloadOptionBase,
    ModelHandleBase,
    ModelInstanceInfo,
    ModelLoadingCallback,
    ModelSessionTypes,
    ModelTypesEmbedding,
    ModelTypesLlm,
    PredictionEndpoint,
    PredictionFirstTokenCallback,
    PredictionFragmentCallback,
    PredictionFragmentEvent,
    PredictionMessageCallback,
    PredictionResult,
    PredictionRoundResult,
    PredictionRxEvent,
    PredictionStreamBase,
    PredictionToolCallEvent,
    PromptProcessingCallback,
    RemoteCallHandler,
    ResponseSchema,
    TModelInfo,
    ToolDefinition,
    check_model_namespace,
    load_struct,
    _model_spec_to_api_dict,
    _redact_json,
)
from ._ws_impl import SyncFutureTimeout
from ._ws_thread import AsyncWebsocketThread, SyncToAsyncWebsocketBridge
from ._kv_config import TLoadConfig, TLoadConfigDict, parse_server_config
from ._sdk_models import (
    EmbeddingRpcCountTokensParameter,
    EmbeddingRpcEmbedStringParameter,
    EmbeddingRpcTokenizeParameter,
    LlmApplyPromptTemplateOpts,
    LlmApplyPromptTemplateOptsDict,
    LlmRpcApplyPromptTemplateParameter,
    ModelCompatibilityType,
)

from ._logging import new_logger, LogEventContext

# Only the sync API itself is published from
# this module. Anything needed for type hints
# and similar tasks is published from `json_api`.
# Bypassing the high level API, and working more
# directly with the underlying websocket(s) is
# supported (hence the public names), but they're
# not exported via the top-level `lmstudio` API.
__all__ = [
    "AnyDownloadedModel",
    "Client",
    "DownloadedEmbeddingModel",
    "DownloadedLlm",
    "EmbeddingModel",
    "LLM",
    "SyncModelHandle",
    "PredictionStream",
    "configure_default_client",
    "get_default_client",
    "get_sync_api_timeout",
    "embedding_model",
    "list_downloaded_models",
    "list_loaded_models",
    "llm",
    "prepare_image",
    "set_sync_api_timeout",
]

#
_DEFAULT_TIMEOUT: float | None = 60.0


@sdk_public_api()
def get_sync_api_timeout() -> float | None:
    """Return the current default sync API timeout when waiting for server messages."""
    return _DEFAULT_TIMEOUT


@sdk_public_api()
def set_sync_api_timeout(timeout: float | None) -> None:
    """Set the default sync API timeout when waiting for server messages."""
    global _DEFAULT_TIMEOUT
    if timeout is not None:
        timeout = float(timeout)
    _DEFAULT_TIMEOUT = timeout


T = TypeVar("T")
CallWithTimeout: TypeAlias = Callable[[float | None], Any]
TimeoutOption: TypeAlias = float | None | EllipsisType


class SyncChannel(Generic[T]):
    """Communication subchannel over multiplexed async websocket."""

    def __init__(
        self,
        channel_id: int,
        get_message: CallWithTimeout,
        endpoint: ChannelEndpoint[T, Any, Any],
        send_json: Callable[[DictObject], None],
        log_context: LogEventContext,
        timeout: TimeoutOption = ...,
    ) -> None:
        """Initialize synchronous websocket streaming channel."""
        self._is_finished = False
        self._get_message = get_message
        self._send_json = send_json
        self._timeout = timeout
        self._api_channel = ChannelHandler(channel_id, endpoint, log_context)

    def get_creation_message(self) -> DictObject:
        """Get the message to send to create this channel."""
        return self._api_channel.get_creation_message()

    def send_message(self, message: DictObject) -> None:
        """Send given message on this channel."""
        wrapped_message = self._api_channel.wrap_message(message)
        self._send_json(wrapped_message)

    def cancel(self) -> None:
        """Cancel the channel."""
        if self._is_finished:
            return
        cancel_message = self._api_channel.get_cancel_message()
        self._send_json(cancel_message)

    @property
    def timeout(self) -> float | None:
        """Permitted time between received messages for this channel."""
        timeout = self._timeout
        if timeout is ...:
            return _DEFAULT_TIMEOUT
        return timeout

    def rx_stream(
        self,
    ) -> Iterator[DictObject | None]:
        """Stream received channel messages until channel is closed by server."""
        while not self._is_finished:
            with sdk_public_api():
                # Avoid emitting tracebacks that delve into supporting libraries
                # (we can't easily suppress the SDK's own frames for iterators)
                try:
                    message = self._get_message(self.timeout)
                except SyncFutureTimeout:
                    raise LMStudioTimeoutError from None
                if message is None:
                    raise LMStudioWebsocketError("Client unexpectedly disconnected.")
                contents = self._api_channel.handle_rx_message(message)
            if contents is None:
                self._is_finished = True
                break
            yield contents

    def wait_for_result(self) -> T:
        """Wait for the channel to finish and return the result."""
        endpoint = self._api_channel.endpoint
        for contents in self.rx_stream():
            endpoint.handle_message_events(contents)
            if endpoint.is_finished:
                break
        return endpoint.result()


class SyncRemoteCall:
    """Remote procedure call over multiplexed async websocket."""

    def __init__(
        self,
        call_id: int,
        get_message: CallWithTimeout,
        log_context: LogEventContext,
        notice_prefix: str = "RPC",
        timeout: TimeoutOption = ...,
    ) -> None:
        """Initialize synchronous remote procedure call."""
        self._get_message = get_message
        self._timeout = timeout
        self._rpc = RemoteCallHandler(call_id, log_context, notice_prefix)
        self._logger = logger = new_logger(type(self).__name__)
        logger.update_context(log_context, call_id=call_id)

    def get_rpc_message(
        self, endpoint: str, params: AnyLMStudioStruct | None
    ) -> DictObject:
        """Get the message to send to initiate this remote procedure call."""
        return self._rpc.get_rpc_message(endpoint, params)

    @property
    def timeout(self) -> float | None:
        """Permitted time to wait for a reply to this call."""
        timeout = self._timeout
        if timeout is ...:
            return _DEFAULT_TIMEOUT
        return timeout

    def receive_result(self) -> Any:
        """Receive call response on the receive queue."""
        try:
            message = self._get_message(self.timeout)
        except SyncFutureTimeout:
            raise LMStudioTimeoutError from None
        if message is None:
            raise LMStudioWebsocketError("Client unexpectedly disconnected.")
        return self._rpc.handle_rx_message(message)


class SyncLMStudioWebsocket(LMStudioWebsocket[SyncToAsyncWebsocketBridge]):
    """Synchronous websocket client that handles demultiplexing of reply messages."""

    def __init__(
        self,
        ws_thread: AsyncWebsocketThread,
        ws_url: str,
        auth_details: DictObject,
        log_context: LogEventContext | None = None,
    ) -> None:
        """Initialize synchronous websocket client."""
        super().__init__(ws_url, auth_details, log_context)
        self._ws_thread = ws_thread

    @property
    def _httpx_ws(self) -> AsyncWebSocketSession | None:
        # Underlying HTTPX session is accessible for testing purposes
        ws_thread = self._ws
        if ws_thread is None:
            return None
        return ws_thread._ws

    def __enter__(self) -> Self:
        # Handle reentrancy the same way files do:
        # allow nested use as a CM, but close on the first exit
        if self._ws is None:
            self.connect()
        return self

    def __exit__(self, *args: Any) -> None:
        self.disconnect()

    def connect(self) -> Self:
        """Connect to and authenticate with the LM Studio API."""
        self._fail_if_connected("Attempted to connect already connected websocket")
        ws = SyncToAsyncWebsocketBridge(
            self._ws_thread,
            self._ws_url,
            self._auth_details,
            self._logger.event_context,
        )
        if not ws.connect():
            if ws._connection_failure is not None:
                raise self._get_connection_failure_error(ws._connection_failure)
            if ws._auth_failure is not None:
                raise self._get_auth_failure_error(ws._auth_failure)
            self._logger.error("Connection failed, but no failure reason reported.")
            raise self._get_connection_failure_error()
        self._ws = ws
        return self

    def disconnect(self) -> None:
        """Drop the LM Studio API connection."""
        ws = self._ws
        self._ws = None
        if ws is not None:
            self._logger.debug(f"Disconnecting websocket session ({self._ws_url})")
            ws.notify_client_termination_threadsafe()
            ws.disconnect()
        self._logger.info(f"Websocket session disconnected ({self._ws_url})")

    close = disconnect

    def _send_json(self, message: DictObject) -> None:
        # Callers are expected to call `_ensure_connected` before this method
        ws = self._ws
        assert ws is not None
        # Background thread handles the exception conversion
        ws.send_json(message)

    def _connect_to_endpoint(self, channel: SyncChannel[Any]) -> None:
        """Connect channel to specified endpoint."""
        self._ensure_connected("open channel endpoints")
        create_message = channel.get_creation_message()
        self._logger.debug("Connecting channel endpoint", json=create_message)
        self._send_json(create_message)

    @contextmanager
    def open_channel(
        self,
        endpoint: ChannelEndpoint[T, Any, Any],
    ) -> Generator[SyncChannel[T], None, None]:
        """Open a streaming channel over the websocket."""
        ws = self._ws
        assert ws is not None
        with ws.open_channel() as (channel_id, getter):
            channel = SyncChannel(
                channel_id,
                getter,
                endpoint,
                self._send_json,
                self._logger.event_context,
            )
            self._connect_to_endpoint(channel)
            yield channel

    def _send_call(
        self,
        rpc: SyncRemoteCall,
        endpoint: str,
        params: AnyLMStudioStruct | None = None,
    ) -> None:
        """Initiate remote call to specified endpoint."""
        self._ensure_connected("send remote procedure call")
        call_message = rpc.get_rpc_message(endpoint, params)
        # TODO: Improve logging for large requests (such as file uploads)
        #       without requiring explicit special casing here
        logged_message: DictObject
        if call_message.get("endpoint") == "uploadFileBase64":
            logged_message = _redact_json(call_message)
        else:
            logged_message = call_message
        self._logger.debug("Sending RPC request", json=logged_message)
        self._send_json(call_message)

    def remote_call(
        self,
        endpoint: str,
        params: AnyLMStudioStruct | None,
        notice_prefix: str = "RPC",
    ) -> Any:
        """Make a remote procedure call over the websocket."""
        ws = self._ws
        assert ws is not None
        with ws.start_call() as (call_id, getter):
            rpc = SyncRemoteCall(
                call_id, getter, self._logger.event_context, notice_prefix
            )
            self._send_call(rpc, endpoint, params)
            return rpc.receive_result()


class _SyncSession(ClientSession["Client", SyncLMStudioWebsocket]):
    """Sync client session interfaces applicable to all API namespaces."""

    def __init__(self, client: "Client") -> None:
        """Initialize synchronous API client session."""
        super().__init__(client)
        self._resources = ExitStack()

    def _ensure_connected(self) -> None:
        # Allow lazy connection of the session websocket
        if self._lmsws is None:
            self.connect()

    def __enter__(self) -> Self:
        # Handle reentrancy the same way files do:
        # allow nested use as a CM, but close on the first exit
        self._ensure_connected()
        return self

    def __exit__(self, *args: Any) -> None:
        self.disconnect()

    @sdk_public_api()
    def connect(self) -> SyncLMStudioWebsocket:
        """Connect the client session."""
        self._fail_if_connected("Attempted to connect already connected session")
        api_host = self._client.api_host
        namespace = self.API_NAMESPACE
        if namespace is None:
            raise LMStudioClientError(
                f"No API namespace defined for {type(self).__name__}"
            )
        session_url = f"ws://{api_host}/{namespace}"
        resources = self._resources
        self._lmsws = lmsws = resources.enter_context(
            SyncLMStudioWebsocket(
                self._client._ws_thread, session_url, self._client._auth_details
            )
        )
        return lmsws

    @sdk_public_api()
    def disconnect(self) -> None:
        """Disconnect the client session."""
        self._lmsws = None
        self._resources.close()

    close = disconnect

    # To allow for client level management of the session lifecycles
    # without requiring network I/O on property access, we implicitly
    # connect the websocket (if necessary) when sending requests

    @contextmanager
    def _create_channel(
        self,
        endpoint: ChannelEndpoint[T, Any, Any],
    ) -> Generator[SyncChannel[T], None, None]:
        """Connect a channel to an LM Studio streaming endpoint."""
        self._ensure_connected()
        lmsws = self._get_lmsws("create channels")
        with lmsws.open_channel(endpoint) as channel:
            yield channel

    @sdk_public_api()
    def remote_call(
        self,
        endpoint: str,
        params: AnyLMStudioStruct | None = None,
        notice_prefix: str = "RPC",
    ) -> Any:
        """Send a remote call to the given RPC endpoint and wait for the result."""
        self._ensure_connected()
        lmsws = self._get_lmsws("make remote calls")
        return lmsws.remote_call(endpoint, params, notice_prefix)


TSyncSessionModel = TypeVar(
    "TSyncSessionModel", bound="_SyncSessionModel[Any, Any, Any, Any]"
)
TModelHandle = TypeVar("TModelHandle", bound="SyncModelHandle[Any]")


class DownloadedModel(
    Generic[
        TModelInfo,
        TSyncSessionModel,
        TLoadConfig,
        TLoadConfigDict,
        TModelHandle,
    ],
    DownloadedModelBase[TModelInfo, TSyncSessionModel],
):
    @sdk_public_api()
    def load_new_instance(
        self,
        *,
        ttl: int | None = DEFAULT_TTL,
        instance_identifier: str | None = None,
        config: TLoadConfig | TLoadConfigDict | None = None,
        on_load_progress: ModelLoadingCallback | None = None,
    ) -> TModelHandle:
        """Load this model with the given identifier and configuration.

        Note: details of configuration fields may change in SDK feature releases.
        """
        handle: TModelHandle = self._session._load_new_instance(
            self.model_key, instance_identifier, ttl, config, on_load_progress
        )
        return handle

    @sdk_public_api()
    def model(
        self,
        *,
        ttl: int | None = DEFAULT_TTL,
        config: TLoadConfig | TLoadConfigDict | None = None,
        on_load_progress: ModelLoadingCallback | None = None,
    ) -> TModelHandle:
        """Retrieve model with default identifier, or load it with given configuration.

        Note: configuration of retrieved model is NOT checked against the given config.
        Note: details of configuration fields may change in SDK feature releases.
        """
        # Call _get_or_load directly, since we have a model identifier
        handle: TModelHandle = self._session._get_or_load(
            self.model_key, ttl, config, on_load_progress
        )
        return handle


class DownloadedEmbeddingModel(
    DownloadedModel[
        EmbeddingModelInfo,
        "_SyncSessionEmbedding",
        EmbeddingLoadModelConfig,
        EmbeddingLoadModelConfigDict,
        "EmbeddingModel",
    ],
):
    """Download listing for an embedding model."""

    def __init__(
        self, model_info: DictObject, session: "_SyncSessionEmbedding"
    ) -> None:
        """Initialize downloaded embedding model details."""
        super().__init__(EmbeddingModelInfo, model_info, session)


class DownloadedLlm(
    DownloadedModel[
        LlmInfo,
        "_SyncSessionLlm",
        LlmLoadModelConfig,
        LlmLoadModelConfigDict,
        "LLM",
    ]
):
    """Download listing for an LLM."""

    def __init__(self, model_info: DictObject, session: "_SyncSessionLlm") -> None:
        """Initialize downloaded embedding model details."""
        super().__init__(LlmInfo, model_info, session)


AnyDownloadedModel: TypeAlias = DownloadedModel[Any, Any, Any, Any, Any]


class _SyncSessionSystem(_SyncSession):
    """Sync client session for the system namespace."""

    API_NAMESPACE = "system"

    @sdk_public_api()
    def list_downloaded_models(self) -> Sequence[AnyDownloadedModel]:
        """Get the list of all downloaded models that are available for loading."""
        # The list of downloaded models is only available via the system API namespace
        models = self.remote_call("listDownloadedModels")
        return [self._process_download_listing(m) for m in models]

    def _process_download_listing(self, model_info: DictObject) -> AnyDownloadedModel:
        model_type = model_info.get("type")
        if model_type is None:
            raise LMStudioClientError(
                f"No 'type' field in download listing: {model_info}"
            )
        match model_type:
            case "embedding":
                return DownloadedEmbeddingModel(model_info, self._client.embedding)
            case "llm":
                return DownloadedLlm(model_info, self._client.llm)
        raise LMStudioClientError(
            f"Unknown model type {model_type!r} in download listing: {model_info}"
        )


class _SyncSessionFiles(_SyncSession):
    """Sync client session for the files namespace."""

    API_NAMESPACE = "files"

    def _fetch_file_handle(self, file_data: _LocalFileData) -> FileHandle:
        handle = self.remote_call("uploadFileBase64", file_data._as_fetch_param())
        # Returned dict provides the handle identifier, file type, and size in bytes
        # Add the extra fields needed for a FileHandle (aka ChatMessagePartFileData)
        handle["name"] = file_data.name
        handle["type"] = "file"
        return load_struct(handle, FileHandle)

    # Not yet implemented (server API only supports the same file types as prepare_image)
    # @sdk_public_api()
    def _prepare_file(self, src: LocalFileInput, name: str | None = None) -> FileHandle:
        """Add a file to the server. Returns a file handle for use in prediction requests."""
        file_data = _LocalFileData(src, name)
        return self._fetch_file_handle(file_data)

    @sdk_public_api()
    def prepare_image(self, src: LocalFileInput, name: str | None = None) -> FileHandle:
        """Add an image to the server. Returns a file handle for use in prediction requests."""
        file_data = _LocalFileData(src, name)
        return self._fetch_file_handle(file_data)


class ModelDownloadOption(ModelDownloadOptionBase[_SyncSession]):
    """A single download option for a model search result."""

    @sdk_public_api()
    def download(
        self,
        on_progress: DownloadProgressCallback | None = None,
        on_finalize: DownloadFinalizedCallback | None = None,
    ) -> str:
        """Download a model and get its path for loading."""
        endpoint = self._get_download_endpoint(on_progress, on_finalize)
        with self._session._create_channel(endpoint) as channel:
            return channel.wait_for_result()


class AvailableModel(AvailableModelBase[_SyncSession]):
    """A model available for download from the model repository."""

    @sdk_public_api()
    def get_download_options(
        self,
    ) -> Sequence[ModelDownloadOption]:
        """Get the download options for the specified model."""
        params = self._get_download_query_params()
        options = self._session.remote_call("getModelDownloadOptions", params)
        final = []
        for m in options["results"]:
            final.append(ModelDownloadOption(m, self._session))
        return final


class _SyncSessionRepository(_SyncSession):
    """Sync client session for the repository namespace."""

    API_NAMESPACE = "repository"

    @sdk_public_api()
    def search_models(
        self,
        search_term: str | None = None,
        limit: int | None = None,
        compatibility_types: list[ModelCompatibilityType] | None = None,
    ) -> Sequence[AvailableModel]:
        """Search for downloadable models satisfying a search query."""
        params = self._get_model_search_params(search_term, limit, compatibility_types)
        models = self.remote_call("searchModels", params)
        return [AvailableModel(m, self) for m in models["results"]]


TDownloadedModel = TypeVar("TDownloadedModel", bound=AnyDownloadedModel)


class _SyncSessionModel(
    _SyncSession,
    Generic[TModelHandle, TLoadConfig, TLoadConfigDict, TDownloadedModel],
):
    """Sync client session for a model (LLM/embedding) namespace."""

    _API_TYPES: Type[ModelSessionTypes[TLoadConfig]]

    @property
    def _system_session(self) -> _SyncSessionSystem:
        return self._client.system

    @property
    def _files_session(self) -> _SyncSessionFiles:
        return self._client.files

    def _get_load_config(self, model_specifier: AnyModelSpecifier) -> AnyLoadConfig:
        """Get the model load config for the specified model."""
        # Note that the configuration reported here uses the *server* config names,
        # not the attributes used to set the configuration in the client SDK
        params = self._API_TYPES.REQUEST_LOAD_CONFIG._from_api_dict(
            {
                "specifier": _model_spec_to_api_dict(model_specifier),
            }
        )
        config = self.remote_call("getLoadConfig", params)
        result_type = self._API_TYPES.MODEL_LOAD_CONFIG
        return result_type._from_any_api_dict(parse_server_config(config))

    def _get_api_model_info(self, model_specifier: AnyModelSpecifier) -> Any:
        """Get the raw model info (if any) for a model matching the given criteria."""
        params = self._API_TYPES.REQUEST_MODEL_INFO._from_api_dict(
            {
                "specifier": _model_spec_to_api_dict(model_specifier),
                "throwIfNotFound": True,
            }
        )
        return self.remote_call("getModelInfo", params)

    @sdk_public_api()
    def get_model_info(self, model_specifier: AnyModelSpecifier) -> ModelInstanceInfo:
        """Get the model info (if any) for a model matching the given criteria."""
        response = self._get_api_model_info(model_specifier)
        model_info = self._API_TYPES.MODEL_INSTANCE_INFO._from_any_api_dict(response)
        return model_info

    def _get_context_length(self, model_specifier: AnyModelSpecifier) -> int:
        """Get the context length of the specified model."""
        raw_model_info = self._get_api_model_info(model_specifier)
        return int(raw_model_info.get("contextLength", -1))

    def _count_tokens(self, model_specifier: AnyModelSpecifier, input: str) -> int:
        params = EmbeddingRpcCountTokensParameter._from_api_dict(
            {
                "specifier": _model_spec_to_api_dict(model_specifier),
                "inputString": input,
            }
        )
        response = self.remote_call("countTokens", params)
        return int(response["tokenCount"])

    # Private helper method to allow the main API to easily accept iterables
    def _tokenize_text(
        self, model_specifier: AnyModelSpecifier, input: str
    ) -> Sequence[int]:
        params = EmbeddingRpcTokenizeParameter._from_api_dict(
            {
                "specifier": _model_spec_to_api_dict(model_specifier),
                "inputString": input,
            }
        )
        response = self.remote_call("tokenize", params)
        return response.get("tokens", []) if response else []

    # Alas, type hints don't properly support distinguishing str vs Iterable[str]:
    #     https://github.com/python/typing/issues/256
    def _tokenize(
        self, model_specifier: AnyModelSpecifier, input: str | Iterable[str]
    ) -> Sequence[int] | Sequence[Sequence[int]]:
        """Tokenize the input string(s) using the specified model."""
        if isinstance(input, str):
            return self._tokenize_text(model_specifier, input)
        return [self._tokenize_text(model_specifier, i) for i in input]

    @abstractmethod
    def _create_handle(self, model_identifier: str) -> TModelHandle:
        """Get a symbolic handle to the specified model."""
        ...

    @sdk_public_api()
    def model(
        self,
        model_key: str | None = None,
        /,
        *,
        ttl: int | None = DEFAULT_TTL,
        config: TLoadConfig | TLoadConfigDict | None = None,
        on_load_progress: ModelLoadingCallback | None = None,
    ) -> TModelHandle:
        """Get a handle to the specified model (loading it if necessary).

        Note: configuration of retrieved model is NOT checked against the given config.
        Note: details of configuration fields may change in SDK feature releases.
        """
        if model_key is None:
            # Should this raise an error if a config is supplied?
            return self._get_any()
        return self._get_or_load(model_key, ttl, config, on_load_progress)

    @sdk_public_api()
    def list_loaded(self) -> Sequence[TModelHandle]:
        """Get the list of currently loaded models."""
        models = self.remote_call("listLoaded")
        return [self._create_handle(m["identifier"]) for m in models]

    @sdk_public_api()
    def unload(self, model_identifier: str) -> None:
        """Unload the specified model."""
        params = self._API_TYPES.REQUEST_UNLOAD(identifier=model_identifier)
        self.remote_call("unloadModel", params)

    # N.B. Canceling a load from the UI doesn't update the load process for a while.
    # Fortunately, this is not our fault. The server just delays in broadcasting it.
    @sdk_public_api()
    def load_new_instance(
        self,
        model_key: str,
        instance_identifier: str | None = None,
        *,
        ttl: int | None = DEFAULT_TTL,
        config: TLoadConfig | TLoadConfigDict | None = None,
        on_load_progress: ModelLoadingCallback | None = None,
    ) -> TModelHandle:
        """Load the specified model with the given identifier and configuration.

        Note: details of configuration fields may change in SDK feature releases.
        """
        return self._load_new_instance(
            model_key, instance_identifier, ttl, config, on_load_progress
        )

    def _load_new_instance(
        self,
        model_key: str,
        instance_identifier: str | None,
        ttl: int | None,
        config: TLoadConfig | TLoadConfigDict | None,
        on_load_progress: ModelLoadingCallback | None,
    ) -> TModelHandle:
        channel_type = self._API_TYPES.REQUEST_NEW_INSTANCE
        config_type = self._API_TYPES.MODEL_LOAD_CONFIG
        endpoint = LoadModelEndpoint(
            model_key,
            instance_identifier,
            ttl,
            channel_type,
            config_type,
            config,
            on_load_progress,
        )
        with self._create_channel(endpoint) as channel:
            result = channel.wait_for_result()
            return self._create_handle(result.identifier)

    def _get_or_load(
        self,
        model_key: str,
        ttl: int | None,
        config: TLoadConfig | TLoadConfigDict | None,
        on_load_progress: ModelLoadingCallback | None,
    ) -> TModelHandle:
        """Get the specified model if it is already loaded, otherwise load it."""
        channel_type = self._API_TYPES.REQUEST_GET_OR_LOAD
        config_type = self._API_TYPES.MODEL_LOAD_CONFIG
        endpoint = GetOrLoadEndpoint(
            model_key, ttl, channel_type, config_type, config, on_load_progress
        )
        with self._create_channel(endpoint) as channel:
            result = channel.wait_for_result()
            return self._create_handle(result.identifier)

    def _get_any(self) -> TModelHandle:
        """Get a handle to any loaded model."""
        loaded_models = self.list_loaded()
        if not loaded_models:
            raise LMStudioClientError(
                f"Could not get model handle in namespace {self.API_NAMESPACE} (no models are currently loaded)."
            )
        return self._create_handle(loaded_models[0].identifier)

    @classmethod
    def _is_relevant_model(cls, model: AnyDownloadedModel) -> TypeIs[TDownloadedModel]:
        return bool(model.type == cls.API_NAMESPACE)

    @sdk_public_api()
    def list_downloaded(self) -> Sequence[TDownloadedModel]:
        """Get the list of currently downloaded models that are available for loading."""
        models = self._system_session.list_downloaded_models()
        return [m for m in models if self._is_relevant_model(m)]

    def _fetch_file_handle(self, file_data: _LocalFileData) -> FileHandle:
        return self._files_session._fetch_file_handle(file_data)


SyncPredictionChannel: TypeAlias = SyncChannel[PredictionResult]
SyncPredictionCM: TypeAlias = ContextManager[SyncPredictionChannel]


class PredictionStream(PredictionStreamBase):
    """Sync context manager for an ongoing prediction process."""

    def __init__(
        self,
        channel_cm: SyncPredictionCM,
        endpoint: PredictionEndpoint,
    ) -> None:
        """Initialize a prediction process representation."""
        self._resources = ExitStack()
        self._channel_cm: SyncPredictionCM = channel_cm
        self._channel: SyncPredictionChannel | None = None
        super().__init__(endpoint)

    @sdk_public_api()
    def start(self) -> None:
        """Send the prediction request."""
        if self._is_finished:
            raise LMStudioRuntimeError("Prediction result has already been received.")
        if self._is_started:
            raise LMStudioRuntimeError("Prediction request has already been sent.")
        # The given channel context manager is set up to send the relevant request
        self._channel = self._resources.enter_context(self._channel_cm)
        self._mark_started()

    @sdk_public_api()
    def close(self) -> None:
        """Terminate the prediction processing (if not already terminated)."""
        # Cancel the prediction (if unfinished) and release acquired resources
        if self._is_started and not self._is_finished:
            self._set_error(
                LMStudioCancelledError(
                    "Prediction cancelled unexpectedly: please use .cancel()"
                )
            )
        self._channel = None
        self._resources.close()

    def __enter__(self) -> Self:
        if self._channel is None:
            self.start()
        return self

    def __exit__(
        self,
        _exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        _exc_tb: TracebackType | None,
    ) -> None:
        if exc_val and not self._is_finished:
            self._set_error(exc_val)
        self.close()

    def __iter__(self) -> Iterator[LlmPredictionFragment]:
        for event in self._iter_events():
            if isinstance(event, PredictionFragmentEvent):
                yield event.arg

    def _iter_events(self) -> Iterator[PredictionRxEvent]:
        endpoint = self._endpoint
        with self:
            assert self._channel is not None
            for contents in self._channel.rx_stream():
                for event in endpoint.iter_message_events(contents):
                    endpoint.handle_rx_event(event)
                    yield event
                if endpoint.is_finished:
                    break
            self._mark_finished()

    @sdk_public_api()
    def wait_for_result(self) -> PredictionResult:
        """Wait for the result of the prediction."""
        for _ in self:
            pass
        return self.result()

    @sdk_public_api()
    def cancel(self) -> None:
        """Cancel the prediction process."""
        if not self._is_finished and self._channel:
            self._mark_cancelled()
            self._channel.cancel()


class _SyncSessionLlm(
    _SyncSessionModel[
        "LLM",
        LlmLoadModelConfig,
        LlmLoadModelConfigDict,
        DownloadedLlm,
    ]
):
    """Sync client session for LLM namespace."""

    API_NAMESPACE = "llm"
    _API_TYPES = ModelTypesLlm

    def __init__(self, client: "Client") -> None:
        """Initialize API client session for LLM interaction."""
        super().__init__(client)

    def _create_handle(self, model_identifier: str) -> "LLM":
        """Create a symbolic handle to the specified LLM model."""
        return LLM(model_identifier, self)

    def _complete_stream(
        self,
        model_specifier: AnyModelSpecifier,
        prompt: str,
        *,
        response_format: ResponseSchema | None = None,
        config: LlmPredictionConfig | LlmPredictionConfigDict | None = None,
        preset: str | None = None,
        on_message: PredictionMessageCallback | None = None,
        on_first_token: PredictionFirstTokenCallback | None = None,
        on_prediction_fragment: PredictionFragmentCallback | None = None,
        on_prompt_processing_progress: PromptProcessingCallback | None = None,
    ) -> PredictionStream:
        """Request a one-off prediction without any context and stream the generated tokens.

        Note: details of configuration fields may change in SDK feature releases.
        """
        endpoint = CompletionEndpoint(
            model_specifier,
            prompt,
            response_format,
            config,
            preset,
            on_message,
            on_first_token,
            on_prediction_fragment,
            on_prompt_processing_progress,
        )
        channel_cm = self._create_channel(endpoint)
        prediction_stream = PredictionStream(channel_cm, endpoint)
        return prediction_stream

    def _respond_stream(
        self,
        model_specifier: AnyModelSpecifier,
        history: Chat | ChatHistoryDataDict | str,
        *,
        response_format: ResponseSchema | None = None,
        config: LlmPredictionConfig | LlmPredictionConfigDict | None = None,
        preset: str | None = None,
        on_message: PredictionMessageCallback | None = None,
        on_first_token: PredictionFirstTokenCallback | None = None,
        on_prediction_fragment: PredictionFragmentCallback | None = None,
        on_prompt_processing_progress: PromptProcessingCallback | None = None,
    ) -> PredictionStream:
        """Request a response in an ongoing assistant chat session and stream the generated tokens.

        Note: details of configuration fields may change in SDK feature releases.
        """
        if not isinstance(history, Chat):
            history = Chat.from_history(history)
        endpoint = ChatResponseEndpoint(
            model_specifier,
            history,
            response_format,
            config,
            preset,
            on_message,
            on_first_token,
            on_prediction_fragment,
            on_prompt_processing_progress,
        )
        channel_cm = self._create_channel(endpoint)
        prediction_stream = PredictionStream(channel_cm, endpoint)
        return prediction_stream

    def _apply_prompt_template(
        self,
        model_specifier: AnyModelSpecifier,
        history: Chat | ChatHistoryDataDict | str,
        opts: LlmApplyPromptTemplateOpts | LlmApplyPromptTemplateOptsDict = {},
    ) -> str:
        """Apply a prompt template to the given history."""
        if not isinstance(history, Chat):
            history = Chat.from_history(history)
        if not isinstance(opts, LlmApplyPromptTemplateOpts):
            opts = LlmApplyPromptTemplateOpts.from_dict(opts)
        params = LlmRpcApplyPromptTemplateParameter._from_api_dict(
            {
                "specifier": _model_spec_to_api_dict(model_specifier),
                "history": history._get_history_for_prediction(),
                "predictionConfigStack": {"layers": []},
                "opts": opts.to_dict(),
            }
        )
        response = self.remote_call("applyPromptTemplate", params)
        return response.get("formatted", "") if response else ""


class _SyncSessionEmbedding(
    _SyncSessionModel[
        "EmbeddingModel",
        EmbeddingLoadModelConfig,
        EmbeddingLoadModelConfigDict,
        DownloadedEmbeddingModel,
    ]
):
    """Sync client session for embedding namespace."""

    API_NAMESPACE = "embedding"
    _API_TYPES = ModelTypesEmbedding

    def __init__(self, client: "Client") -> None:
        """Initialize API client session for embedding model interaction."""
        super().__init__(client)

    def _create_handle(self, model_identifier: str) -> "EmbeddingModel":
        """Create a symbolic handle to the specified embedding model."""
        return EmbeddingModel(model_identifier, self)

    # Private helper method to allow the main API to easily accept iterables
    def _embed_text(
        self, model_specifier: AnyModelSpecifier, input: str
    ) -> Sequence[float]:
        params = EmbeddingRpcEmbedStringParameter._from_api_dict(
            {
                "modelSpecifier": _model_spec_to_api_dict(model_specifier),
                "inputString": input,
            }
        )

        response = self.remote_call("embedString", params)
        return response.get("embedding", []) if response else []

    # Alas, type hints don't properly support distinguishing str vs Iterable[str]:
    #     https://github.com/python/typing/issues/256
    def _embed(
        self, model_specifier: AnyModelSpecifier, input: str | Iterable[str]
    ) -> Sequence[float] | Sequence[Sequence[float]]:
        """Request embedding vectors for the given input string(s)."""
        if isinstance(input, str):
            return self._embed_text(model_specifier, input)
        return [self._embed_text(model_specifier, i) for i in input]


class SyncModelHandle(ModelHandleBase[TSyncSessionModel]):
    """Reference to a loaded LM Studio model."""

    @sdk_public_api()
    def unload(self) -> None:
        """Unload this model."""
        self._session.unload(self.identifier)

    @sdk_public_api()
    def get_info(self) -> ModelInstanceInfo:
        """Get the model info for this model."""
        return self._session.get_model_info(self.identifier)

    @sdk_public_api()
    def get_load_config(self) -> AnyLoadConfig:
        """Get the model load config for this model."""
        return self._session._get_load_config(self.identifier)

    # Alas, type hints don't properly support distinguishing str vs Iterable[str]:
    #     https://github.com/python/typing/issues/256
    @sdk_public_api()
    def tokenize(
        self, input: str | Iterable[str]
    ) -> Sequence[int] | Sequence[Sequence[int]]:
        """Tokenize the input string(s) using this model."""
        return self._session._tokenize(self.identifier, input)

    @sdk_public_api()
    def count_tokens(self, input: str) -> int:
        """Report the number of tokens needed for the input string using this model."""
        return self._session._count_tokens(self.identifier, input)

    @sdk_public_api()
    def get_context_length(self) -> int:
        """Get the context length of this model."""
        return self._session._get_context_length(self.identifier)


AnySyncModel: TypeAlias = SyncModelHandle[Any]


class LLM(SyncModelHandle[_SyncSessionLlm]):
    """Reference to a loaded LLM model."""

    @sdk_public_api()
    def complete_stream(
        self,
        prompt: str,
        *,
        response_format: ResponseSchema | None = None,
        config: LlmPredictionConfig | LlmPredictionConfigDict | None = None,
        preset: str | None = None,
        on_message: PredictionMessageCallback | None = None,
        on_first_token: PredictionFirstTokenCallback | None = None,
        on_prediction_fragment: PredictionFragmentCallback | None = None,
        on_prompt_processing_progress: PromptProcessingCallback | None = None,
    ) -> PredictionStream:
        """Request a one-off prediction without any context and stream the generated tokens.

        Note: details of configuration fields may change in SDK feature releases.
        """
        return self._session._complete_stream(
            self.identifier,
            prompt,
            response_format=response_format,
            config=config,
            preset=preset,
            on_message=on_message,
            on_first_token=on_first_token,
            on_prediction_fragment=on_prediction_fragment,
            on_prompt_processing_progress=on_prompt_processing_progress,
        )

    @sdk_public_api()
    def complete(
        self,
        prompt: str,
        *,
        response_format: ResponseSchema | None = None,
        config: LlmPredictionConfig | LlmPredictionConfigDict | None = None,
        preset: str | None = None,
        on_message: PredictionMessageCallback | None = None,
        on_first_token: PredictionFirstTokenCallback | None = None,
        on_prediction_fragment: PredictionFragmentCallback | None = None,
        on_prompt_processing_progress: PromptProcessingCallback | None = None,
    ) -> PredictionResult:
        """Request a one-off prediction without any context.

        Note: details of configuration fields may change in SDK feature releases.
        """
        prediction_stream = self._session._complete_stream(
            self.identifier,
            prompt,
            response_format=response_format,
            config=config,
            preset=preset,
            on_message=on_message,
            on_first_token=on_first_token,
            on_prediction_fragment=on_prediction_fragment,
            on_prompt_processing_progress=on_prompt_processing_progress,
        )
        for _ in prediction_stream:
            # No yield in body means iterator reliably provides
            # prompt resource cleanup on coroutine cancellation
            pass
        return prediction_stream.result()

    @sdk_public_api()
    def respond_stream(
        self,
        history: Chat | ChatHistoryDataDict | str,
        *,
        response_format: ResponseSchema | None = None,
        config: LlmPredictionConfig | LlmPredictionConfigDict | None = None,
        preset: str | None = None,
        on_message: PredictionMessageCallback | None = None,
        on_first_token: PredictionFirstTokenCallback | None = None,
        on_prediction_fragment: PredictionFragmentCallback | None = None,
        on_prompt_processing_progress: PromptProcessingCallback | None = None,
    ) -> PredictionStream:
        """Request a response in an ongoing assistant chat session and stream the generated tokens.

        Note: details of configuration fields may change in SDK feature releases.
        """
        return self._session._respond_stream(
            self.identifier,
            history,
            response_format=response_format,
            config=config,
            preset=preset,
            on_message=on_message,
            on_first_token=on_first_token,
            on_prediction_fragment=on_prediction_fragment,
            on_prompt_processing_progress=on_prompt_processing_progress,
        )

    @sdk_public_api()
    def respond(
        self,
        history: Chat | ChatHistoryDataDict | str,
        *,
        response_format: ResponseSchema | None = None,
        config: LlmPredictionConfig | LlmPredictionConfigDict | None = None,
        preset: str | None = None,
        on_message: PredictionMessageCallback | None = None,
        on_first_token: PredictionFirstTokenCallback | None = None,
        on_prediction_fragment: PredictionFragmentCallback | None = None,
        on_prompt_processing_progress: PromptProcessingCallback | None = None,
    ) -> PredictionResult:
        """Request a response in an ongoing assistant chat session.

        Note: details of configuration fields may change in SDK feature releases.
        """
        prediction_stream = self._session._respond_stream(
            self.identifier,
            history,
            response_format=response_format,
            config=config,
            preset=preset,
            on_message=on_message,
            on_first_token=on_first_token,
            on_prediction_fragment=on_prediction_fragment,
            on_prompt_processing_progress=on_prompt_processing_progress,
        )
        for _ in prediction_stream:
            # No yield in body means iterator reliably provides
            # prompt resource cleanup on coroutine cancellation
            pass
        return prediction_stream.result()

    # TODO: Improve code sharing between sync and async multi-round predictions
    # TODO: Accept async tools in the tools iterable
    @sdk_public_api()
    def act(
        self,
        chat: Chat | ChatHistoryDataDict | str,
        tools: Iterable[ToolDefinition],
        *,
        max_prediction_rounds: int | None = None,
        max_parallel_tool_calls: int | None = 1,
        config: LlmPredictionConfig | LlmPredictionConfigDict | None = None,
        preset: str | None = None,
        on_message: Callable[[AssistantResponse | ToolResultMessage], Any]
        | None = None,
        on_first_token: Callable[[int], Any] | None = None,
        on_prediction_fragment: Callable[[LlmPredictionFragment, int], Any]
        | None = None,
        on_round_start: Callable[[int], Any] | None = None,
        on_round_end: Callable[[int], Any] | None = None,
        on_prediction_completed: Callable[[PredictionRoundResult], Any] | None = None,
        on_prompt_processing_progress: Callable[[float, int], Any] | None = None,
        handle_invalid_tool_request: Callable[
            [LMStudioPredictionError, ToolCallRequest | None], str | None
        ]
        | None = None,
    ) -> ActResult:
        """Request a response (with implicit tool use) in an ongoing agent chat session.

        Note: details of configuration fields may change in SDK feature releases.
        """
        start_time = time.perf_counter()
        # It is not yet possible to combine tool calling with requests for structured responses
        response_format = None
        agent_chat: Chat = Chat.from_history(chat)
        del chat  # Avoid any further access to the input chat history
        # Multiple rounds, until all tool calls are resolved or limit is reached
        round_counter: Iterable[int]
        if max_prediction_rounds is not None:
            if max_prediction_rounds < 1:
                raise LMStudioValueError(
                    f"Max prediction rounds must be at least 1 ({max_prediction_rounds!r} given)"
                )
            round_counter = range(max_prediction_rounds)
            final_round_index = max_prediction_rounds - 1
        else:
            # Do not force a final round when no limit is specified
            final_round_index = -1
            round_counter = itertools.count()
        llm_tool_args = ChatResponseEndpoint.parse_tools(tools)
        del tools
        # Supply the round index to any endpoint callbacks that expect one
        round_index: int
        on_first_token_for_endpoint: PredictionFirstTokenCallback | None = None
        if on_first_token is not None:

            def _wrapped_on_first_token() -> None:
                assert on_first_token is not None
                on_first_token(round_index)

            on_first_token_for_endpoint = _wrapped_on_first_token
        on_prediction_fragment_for_endpoint: PredictionFragmentCallback | None = None
        if on_prediction_fragment is not None:

            def _wrapped_on_prediction_fragment(
                fragment: LlmPredictionFragment,
            ) -> None:
                assert on_prediction_fragment is not None
                on_prediction_fragment(fragment, round_index)

            on_prediction_fragment_for_endpoint = _wrapped_on_prediction_fragment
        on_prompt_processing_for_endpoint: PromptProcessingCallback | None = None
        if on_prompt_processing_progress is not None:

            def _wrapped_on_prompt_processing_progress(progress: float) -> None:
                assert on_prompt_processing_progress is not None
                on_prompt_processing_progress(progress, round_index)

            on_prompt_processing_for_endpoint = _wrapped_on_prompt_processing_progress
        # TODO: Implementation to this point is common between the sync and async APIs
        # (aside from the allow_async flag when parsing the tool definitions)
        # Implementation past this point differs (as the async API uses the loop's executor)

        # Request predictions until no more tool call requests are received in response
        # (or the maximum number of prediction rounds is reached)
        with ThreadPoolExecutor(max_parallel_tool_calls) as pool:
            for round_index in round_counter:
                self._logger.debug(
                    "Starting .act() prediction round", round_index=round_index
                )
                if on_round_start is not None:
                    err_msg = f"Round start callback failed for {self!r}"
                    with sdk_callback_invocation(err_msg, self._logger):
                        on_round_start(round_index)
                # Update the endpoint definition on each iteration in order to:
                # * update the chat history with the previous round result
                # * be able to disallow tool use when the rounds are limited
                # TODO: Refactor endpoint API to avoid repeatedly performing the
                #       LlmPredictionConfig -> KvConfigStack transformation
                endpoint = ChatResponseEndpoint(
                    self.identifier,
                    agent_chat,
                    response_format,
                    config,
                    preset,
                    None,  # on_message is invoked directly
                    on_first_token_for_endpoint,
                    on_prediction_fragment_for_endpoint,
                    on_prompt_processing_for_endpoint,
                    handle_invalid_tool_request,
                    *(
                        llm_tool_args
                        if round_index != final_round_index
                        else (None, None)
                    ),
                )
                channel_cm = self._session._create_channel(endpoint)
                prediction_stream = PredictionStream(channel_cm, endpoint)
                tool_call_requests: list[ToolCallRequest] = []
                pending_tool_calls: dict[SyncFuture[Any], ToolCallRequest] = {}
                for event in prediction_stream._iter_events():
                    if isinstance(event, PredictionToolCallEvent):
                        tool_call_request = event.arg
                        tool_call_requests.append(tool_call_request)
                        # TODO: Also handle async tool calls here
                        tool_call = endpoint.request_tool_call(tool_call_request)
                        pending_tool_calls[pool.submit(tool_call)] = tool_call_request
                prediction = prediction_stream.result()
                self._logger.debug(
                    "Completed .act() prediction round", round_index=round_index
                )
                if on_prediction_completed:
                    round_result = PredictionRoundResult.from_result(
                        prediction, round_index
                    )
                    err_msg = f"Prediction completed callback failed for {self!r}"
                    with sdk_callback_invocation(err_msg, self._logger):
                        on_prediction_completed(round_result)
                if pending_tool_calls:

                    def _finish_tool_call(fut: SyncFuture[Any]) -> Any:
                        exc = fut.exception()
                        if exc is not None:
                            if not isinstance(exc, Exception):
                                # Don't allow base exceptions to be suppressed
                                raise exc
                            failed_request = pending_tool_calls[fut]
                            return endpoint._handle_failed_tool_request(
                                exc, failed_request
                            )
                        return fut.result()

                    tool_results = [
                        _finish_tool_call(fut)
                        for fut in as_completed(pending_tool_calls)
                    ]
                    requests_message = agent_chat.add_assistant_response(
                        prediction, tool_call_requests
                    )
                    results_message = agent_chat.add_tool_results(tool_results)
                    if on_message is not None:
                        err_msg = f"Tool request message callback failed for {self!r}"
                        with sdk_callback_invocation(err_msg, self._logger):
                            on_message(requests_message)
                        err_msg = f"Tool result message callback failed for {self!r}"
                        with sdk_callback_invocation(err_msg, self._logger):
                            on_message(results_message)
                elif on_message is not None:
                    err_msg = f"Final response message callback failed for {self!r}"
                    with sdk_callback_invocation(err_msg, self._logger):
                        on_message(agent_chat.add_assistant_response(prediction))
                if on_round_end is not None:
                    err_msg = f"Round end callback failed for {self!r}"
                    with sdk_callback_invocation(err_msg, self._logger):
                        on_round_end(round_index)
                if not tool_call_requests:
                    # No tool call requests -> we're done here
                    break
                if round_index == final_round_index:
                    # We somehow received at least one tool call request,
                    # even though tools are omitted on the final round
                    err_msg = "Model requested tool use on final prediction round."
                    endpoint._handle_invalid_tool_request(err_msg)
                    break
        num_rounds = round_index + 1
        duration = time.perf_counter() - start_time
        return ActResult(rounds=num_rounds, total_time_seconds=duration)

    @sdk_public_api()
    def apply_prompt_template(
        self,
        history: Chat | ChatHistoryDataDict | str,
        opts: LlmApplyPromptTemplateOpts | LlmApplyPromptTemplateOptsDict = {},
    ) -> str:
        """Apply a prompt template to the given history."""
        return self._session._apply_prompt_template(
            self.identifier,
            history,
            opts=opts,
        )


class EmbeddingModel(SyncModelHandle[_SyncSessionEmbedding]):
    """Reference to a loaded embedding model."""

    # Alas, type hints don't properly support distinguishing str vs Iterable[str]:
    #     https://github.com/python/typing/issues/256
    @sdk_public_api()
    def embed(
        self, input: str | Iterable[str]
    ) -> Sequence[float] | Sequence[Sequence[float]]:
        """Request embedding vectors for the given input string(s)."""
        return self._session._embed(self.identifier, input)


TSyncSession = TypeVar("TSyncSession", bound=_SyncSession)


class Client(ClientBase):
    """Synchronous SDK client interface."""

    def __init__(self, api_host: str | None = None) -> None:
        """Initialize API client."""
        super().__init__(api_host)
        self._resources = rm = ExitStack()
        self._ws_thread = ws_thread = AsyncWebsocketThread(dict(client=repr(self)))
        ws_thread.start()
        rm.callback(ws_thread.terminate)
        self._sessions: dict[str, _SyncSession] = {}
        # Support GC-based resource management in the sync API by
        # finalizing at the client layer, and letting its resource
        # manager handle clearing up everything else
        rm.callback(self._sessions.clear)
        weakref.finalize(self, rm.close)

    def __enter__(self) -> Self:
        # Handle reentrancy the same way files do:
        # allow nested use as a CM, but close on the first exit
        with sdk_public_api():
            self._ensure_api_host_is_valid()
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()

    def close(self) -> None:
        """Close any started client sessions."""
        self._resources.close()

    @staticmethod
    def _query_probe_url(url: str) -> httpx.Response:
        return httpx.get(url, timeout=1)

    @classmethod
    @sdk_public_api()
    def is_valid_api_host(cls, api_host: str) -> bool:
        """Report whether the given API host is running an API server instance."""
        probe_url = cls._get_probe_url(api_host)
        try:
            probe_response = cls._query_probe_url(probe_url)
        except (httpx.ConnectTimeout, httpx.ConnectError):
            return False
        return cls._check_probe_response(probe_response)

    @classmethod
    @sdk_public_api()
    def find_default_local_api_host(cls) -> str | None:
        """Query local ports for a running API server instance."""
        for api_host in cls._iter_default_api_hosts():
            if cls.is_valid_api_host(api_host):
                return api_host
        return None

    def _ensure_api_host_is_valid(self) -> None:
        specified_api_host = self._api_host
        if specified_api_host is None:
            api_host = self.find_default_local_api_host()
        elif self.is_valid_api_host(specified_api_host):
            api_host = specified_api_host
        else:
            api_host = None
        if api_host is None:
            raise self._get_probe_failure_error(specified_api_host)
        self._api_host = api_host

    # Doing network I/O in properties is generally considered undesirable.
    # The async API can't perform network I/O in properties at all.
    # Unlike the async API (which follows the principles of structured
    # concurrency), the sync API can resolve this by delaying the creation
    # of the underlying websockets until they're actually used.

    def _get_session(self, cls: Type[TSyncSession]) -> TSyncSession:
        """Get the client session of the given type."""
        namespace = cls.API_NAMESPACE
        assert namespace is not None
        try:
            session = self._sessions[namespace]
        except KeyError:
            pass
        else:
            # This *will* be an instance of the given type.
            # The assertion notifies typecheckers of that.
            assert isinstance(session, cls)
            return session
        # No session yet for this namespace, so create one
        session = cls(self)
        self._sessions[namespace] = session
        # Note: session starts itself on first use rather than immediately
        # We push the exit callback here so that only the client itself
        # needs to be explicitly managed as a context manager. Directly
        # managing the individual session is supported, but not required.
        self._resources.push(session)
        return session

    @property
    @sdk_public_api()
    def llm(self) -> _SyncSessionLlm:
        """Return the LLM API client session."""
        return self._get_session(_SyncSessionLlm)

    @property
    @sdk_public_api()
    def embedding(self) -> _SyncSessionEmbedding:
        """Return the embedding model API client session."""
        return self._get_session(_SyncSessionEmbedding)

    @property
    def system(self) -> _SyncSessionSystem:
        """Return the system API client session."""
        return self._get_session(_SyncSessionSystem)

    @property
    def files(self) -> _SyncSessionFiles:
        """Return the files API client session."""
        return self._get_session(_SyncSessionFiles)

    @property
    def repository(self) -> _SyncSessionRepository:
        """Return the repository API client session."""
        return self._get_session(_SyncSessionRepository)

    # Convenience methods
    # Not yet implemented (server API only supports the same file types as prepare_image)
    # @sdk_public_api()
    def _prepare_file(self, src: LocalFileInput, name: str | None = None) -> FileHandle:
        """Add a file to the server. Returns a file handle for use in prediction requests."""
        return self.files._prepare_file(src, name)

    @sdk_public_api()
    def prepare_image(self, src: LocalFileInput, name: str | None = None) -> FileHandle:
        """Add an image to the server. Returns a file handle for use in prediction requests."""
        return self.files.prepare_image(src, name)

    @sdk_public_api()
    def list_downloaded_models(
        self, namespace: str | None = None
    ) -> Sequence[AnyDownloadedModel]:
        """Get the list of downloaded models."""
        namespace_filter = check_model_namespace(namespace)
        if namespace_filter is None:
            return self.system.list_downloaded_models()
        if namespace_filter == "llm":
            return self.llm.list_downloaded()
        return self.embedding.list_downloaded()

    @sdk_public_api()
    def list_loaded_models(
        self, namespace: str | None = None
    ) -> Sequence[AnySyncModel]:
        """Get the list of loaded models using the default global client."""
        namespace_filter = check_model_namespace(namespace)
        loaded_models: list[AnySyncModel] = []
        if namespace_filter is None or namespace_filter == "llm":
            loaded_models.extend(self.llm.list_loaded())
        if namespace_filter is None or namespace_filter == "embedding":
            loaded_models.extend(self.embedding.list_loaded())
        return loaded_models


# Convenience API
_default_api_host = None
_default_client: Client | None = None


@sdk_public_api()
def configure_default_client(api_host: str) -> None:
    """Set the server API host for the default global client (without creating the client)."""
    global _default_api_host
    if _default_client is not None:
        raise LMStudioClientError(
            "Default client is already created, cannot set its API host."
        )
    _default_api_host = api_host


@sdk_public_api()
def get_default_client(api_host: str | None = None) -> Client:
    """Get the default global client (creating it if necessary)."""
    global _default_client
    if api_host is not None:
        # This will raise an exception if the client already exists
        configure_default_client(api_host)
    if _default_client is None:
        _default_client = Client(_default_api_host)
        _default_client._ensure_api_host_is_valid()
    return _default_client


def _reset_default_client() -> None:
    # Allow the test suite to reset the client without
    # having to poke directly at the module's internals
    global _default_api_host, _default_client
    previous_client = _default_client
    _default_api_host = _default_client = None
    if previous_client is not None:
        previous_client.close()


@sdk_public_api()
def llm(
    model_key: str | None = None,
    /,
    *,
    ttl: int | None = DEFAULT_TTL,
    config: LlmLoadModelConfig | LlmLoadModelConfigDict | None = None,
) -> LLM:
    """Access an LLM using the default global client.

    Note: configuration of retrieved model is NOT checked against the given config.
    Note: details of configuration fields may change in SDK feature releases.
    """
    return get_default_client().llm.model(model_key, ttl=ttl, config=config)


@sdk_public_api()
def embedding_model(
    model_key: str | None = None,
    /,
    *,
    ttl: int | None = DEFAULT_TTL,
    config: EmbeddingLoadModelConfig | EmbeddingLoadModelConfigDict | None = None,
) -> EmbeddingModel:
    """Access an embedding model using the default global client.

    Note: configuration of retrieved model is NOT checked against the given config.
    Note: details of configuration fields may change in SDK feature releases.
    """
    return get_default_client().embedding.model(model_key, ttl=ttl, config=config)


# Not yet implemented (server API only supports the same file types as prepare_image)
# @sdk_public_api()
def _prepare_file(src: LocalFileInput, name: str | None = None) -> FileHandle:
    """Add a file to the server using the default global client."""
    return get_default_client()._prepare_file(src, name)


@sdk_public_api()
def prepare_image(src: LocalFileInput, name: str | None = None) -> FileHandle:
    """Add an image to the server using the default global client."""
    return get_default_client().prepare_image(src, name)


@sdk_public_api()
def list_downloaded_models(
    namespace: str | None = None,
) -> Sequence[AnyDownloadedModel]:
    """Get the list of downloaded models using the default global client."""
    return get_default_client().list_downloaded_models(namespace)


@sdk_public_api()
def list_loaded_models(namespace: str | None = None) -> Sequence[AnySyncModel]:
    """Get the list of loaded models using the default global client."""
    return get_default_client().list_loaded_models(namespace)



#####################################################################
# This files has been automatically generated from:
#   ../async/test_llm_async.py
#
# DO NOT EDIT THIS FILE! Edit the async test case listed above,
# and regenerate the synchronous test cases with async2sync.py
#####################################################################
"""Test non-inference methods on LLMs."""

import logging
from contextlib import nullcontext

import pytest
from pytest import LogCaptureFixture as LogCap

from lmstudio import (
    Client,
    LlmLoadModelConfig,
    LMStudioModelNotFoundError,
    history,
)

from ..support import EXPECTED_LLM, EXPECTED_LLM_ID, check_sdk_error


@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_LLM, EXPECTED_LLM_ID))
def test_apply_prompt_template_sync(model_id: str, caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    file_data = history.FileHandle(
        name="someFile.txt",
        identifier="some-file",
        size_bytes=100,
        file_type="text/plain",
    )
    other_file_data = {
        "name": "someOtherFile.txt",
        "identifier": "some-other-file",
        "sizeBytes": 100,
        # Ensure snake_case field is normalised before sending
        "file_type": "text/plain",
    }
    context = history.Chat("Initial system messages")
    context.add_user_message("Simple text prompt")
    context.add_user_message(history.TextData(text="Structured text prompt"))
    context.add_user_message(file_data)
    context.add_user_message(other_file_data)
    context.add_system_prompt("Simple text prompt")
    context.add_assistant_response("Simple text response")
    accumulated_history = context._get_history_for_prediction()
    with Client() as client:
        templated = client.llm._apply_prompt_template(model_id, accumulated_history)
    logging.info(f"Prompt template: {templated}")
    assert templated


@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_LLM, EXPECTED_LLM_ID))
def test_tokenize_sync(model_id: str, caplog: LogCap) -> None:
    text = "Hello, world!"

    caplog.set_level(logging.DEBUG)
    with Client() as client:
        model = client.llm.model(model_id)
        num_tokens = model.count_tokens(text)
        response = model.tokenize(text)
    logging.info(f"Tokenization response: {response}")
    assert response
    assert isinstance(response, list)
    # Ensure token count and tokenization are consistent
    assert len(response) == num_tokens


@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_LLM, EXPECTED_LLM_ID))
def test_tokenize_list_sync(model_id: str, caplog: LogCap) -> None:
    text = ["Hello, world!", "Goodbye, world!"]

    caplog.set_level(logging.DEBUG)
    with Client() as client:
        model = client.llm.model(model_id)
        response = model.tokenize(text)
    logging.info(f"Tokenization response: {response}")
    assert response
    assert isinstance(response, list)
    assert len(response) == len(text)
    assert all(isinstance(tokens, list) for tokens in response)


@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_LLM, EXPECTED_LLM_ID))
def test_context_length_sync(model_id: str, caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        response = client.llm._get_context_length(model_id)
    logging.info(f"Context length response: {response}")
    assert response
    assert isinstance(response, int)
    assert response > 0


@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_LLM, EXPECTED_LLM_ID))
def test_get_load_config_sync(model_id: str, caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        model = client.llm.model(model_id)
        response = model.get_load_config()
    logging.info(f"Load config response: {response}")
    assert response
    assert isinstance(response, LlmLoadModelConfig)


@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_LLM, EXPECTED_LLM_ID))
def test_get_model_info_sync(model_id: str, caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        response = client.llm.get_model_info(model_id)
    logging.info(f"Model config response: {response}")
    assert response


@pytest.mark.lmstudio
def test_invalid_model_request_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        # Deliberately create an invalid model handle
        model = client.llm._create_handle("No such model")
        # This should error rather than timing out,
        # but avoid any risk of the client hanging...
        with nullcontext():
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                model.complete("Some text")
            check_sdk_error(exc_info, __file__)
        with nullcontext():
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                model.respond("Some text")
            check_sdk_error(exc_info, __file__)
        with nullcontext():
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                model.count_tokens("Some text")
        with nullcontext():
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                model.tokenize("Some text")
            check_sdk_error(exc_info, __file__)
        with nullcontext():
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                model.get_context_length()
            check_sdk_error(exc_info, __file__)



"""Test embedding strings with the API."""

import logging

import anyio
import pytest
from pytest import LogCaptureFixture as LogCap

from lmstudio import AsyncClient, EmbeddingLoadModelConfig, LMStudioModelNotFoundError

from ..support import (
    EXPECTED_EMBEDDING,
    EXPECTED_EMBEDDING_CONTEXT_LENGTH,
    EXPECTED_EMBEDDING_ID,
    EXPECTED_EMBEDDING_LENGTH,
    check_sdk_error,
)


@pytest.mark.asyncio
@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_EMBEDDING, EXPECTED_EMBEDDING_ID))
async def test_embedding_async(model_id: str, caplog: LogCap) -> None:
    text = "Hello, world!"

    caplog.set_level(logging.DEBUG)
    async with AsyncClient() as client:
        session = client.embedding
        response = await session._embed(model_id, input=text)
    logging.info(f"Embedding response: {response}")
    assert response
    assert isinstance(response, list)
    assert len(response) == EXPECTED_EMBEDDING_LENGTH
    # the response should be deterministic if we set constant seed
    # so we can also check the value if desired


@pytest.mark.asyncio
@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_EMBEDDING, EXPECTED_EMBEDDING_ID))
async def test_embedding_list_async(model_id: str, caplog: LogCap) -> None:
    text = ["Hello, world!", "Goodbye, world!"]

    caplog.set_level(logging.DEBUG)
    async with AsyncClient() as client:
        session = client.embedding
        response = await session._embed(model_id, input=text)
    logging.info(f"Embedding response: {response}")
    assert response
    assert isinstance(response, list)
    assert len(response) == len(text)
    assert all(isinstance(embed, list) for embed in response)
    assert all(len(embed) == EXPECTED_EMBEDDING_LENGTH for embed in response)
    # the response should be deterministic if we set constant seed
    # so we can also check the value if desired


@pytest.mark.asyncio
@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_EMBEDDING, EXPECTED_EMBEDDING_ID))
async def test_tokenize_async(model_id: str, caplog: LogCap) -> None:
    text = "Hello, world!"

    caplog.set_level(logging.DEBUG)
    async with AsyncClient() as client:
        model = await client.embedding.model(model_id)
        num_tokens = await model.count_tokens(text)
        response = await model.tokenize(text)
    logging.info(f"Tokenization response: {response}")
    assert response
    assert isinstance(response, list)
    # Ensure token count and tokenization are consistent
    # (embedding models add extra start/end markers during actual tokenization)
    assert len(response) == num_tokens + 2
    # the response should be deterministic if we set constant seed
    # so we can also check the value if desired


@pytest.mark.asyncio
@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_EMBEDDING, EXPECTED_EMBEDDING_ID))
async def test_tokenize_list_async(model_id: str, caplog: LogCap) -> None:
    text = ["Hello, world!", "Goodbye, world!"]

    caplog.set_level(logging.DEBUG)
    async with AsyncClient() as client:
        model = await client.embedding.model(model_id)
        response = await model.tokenize(text)
    logging.info(f"Tokenization response: {response}")
    assert response
    assert isinstance(response, list)
    assert len(response) == len(text)
    assert all(isinstance(tokens, list) for tokens in response)
    # the response should be deterministic if we set constant seed
    # so we can also check the value if desired


@pytest.mark.asyncio
@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_EMBEDDING, EXPECTED_EMBEDDING_ID))
async def test_context_length_async(model_id: str, caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    async with AsyncClient() as client:
        session = client.embedding
        response = await session._get_context_length(model_id)
    logging.info(f"Context length response: {response}")
    assert response
    assert isinstance(response, int)
    assert response == EXPECTED_EMBEDDING_CONTEXT_LENGTH


@pytest.mark.asyncio
@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_EMBEDDING, EXPECTED_EMBEDDING_ID))
async def test_get_load_config_async(model_id: str, caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    async with AsyncClient() as client:
        model = await client.embedding.model(model_id)
        response = await model.get_load_config()
    logging.info(f"Load config response: {response}")
    assert response
    assert isinstance(response, EmbeddingLoadModelConfig)


@pytest.mark.asyncio
@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_EMBEDDING, EXPECTED_EMBEDDING_ID))
async def test_get_model_info_async(model_id: str, caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    async with AsyncClient() as client:
        response = await client.embedding.get_model_info(model_id)
    logging.info(f"Model config response: {response}")
    assert response


@pytest.mark.asyncio
@pytest.mark.lmstudio
async def test_invalid_model_request_async(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    async with AsyncClient() as client:
        # Deliberately create an invalid model handle
        model = client.embedding._create_handle("No such model")
        # This should error rather than timing out,
        # but avoid any risk of the client hanging...
        with anyio.fail_after(30):
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                await model.embed("Some text")
            check_sdk_error(exc_info, __file__)
        with anyio.fail_after(30):
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                await model.count_tokens("Some text")
            check_sdk_error(exc_info, __file__)
        with anyio.fail_after(30):
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                await model.tokenize("Some text")
            check_sdk_error(exc_info, __file__)
        with anyio.fail_after(30):
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                await model.get_context_length()
            check_sdk_error(exc_info, __file__)



"""Invoking and supporting token generator hook implementations."""

from typing import Any, Awaitable, Callable
from ..._sdk_models import (
    # TODO: Define aliases at schema generation time
    PluginsChannelSetGeneratorToClientPacketGenerate as TokenGenerationRequest,
)

from ..config_schemas import BaseConfigSchema
from .common import (
    _AsyncSessionPlugins,
    HookController,
    TPluginConfigSchema,
    TGlobalConfigSchema,
)

# Available as lmstudio.plugin.hooks.*
__all__ = [
    "TokenGeneratorController",
    "TokenGeneratorHook",
    "run_token_generator",
]


class TokenGeneratorController(
    HookController[TokenGenerationRequest, TPluginConfigSchema, TGlobalConfigSchema]
):
    """API access for token generator hook implementations."""


TokenGeneratorHook = Callable[[TokenGeneratorController[Any, Any]], Awaitable[None]]


async def run_token_generator(
    plugin_name: str,
    hook_impl: TokenGeneratorHook,
    plugin_config_schema: type[BaseConfigSchema],
    global_config_schema: type[BaseConfigSchema],
    session: _AsyncSessionPlugins,
    notify_ready: Callable[[], Any],
) -> None:
    """Accept token generation requests."""
    raise NotImplementedError



"""Test plugin config schema definitions."""

from lmstudio.plugin import BaseConfigSchema, config_field


def test_empty_config() -> None:
    class ConfigSchema(BaseConfigSchema):
        pass

    assert ConfigSchema._to_kv_config_schematics() is None


def test_config_field_bool() -> None:
    class ConfigSchema(BaseConfigSchema):
        setting: bool = config_field(label="UI label", hint="UI tooltip", default=True)

    assert ConfigSchema.setting is True
    assert ConfigSchema().setting is True
    kv_config_schematics = ConfigSchema._to_kv_config_schematics()
    assert kv_config_schematics is not None
    expected_kv_config_schematics = {
        "fields": [
            {
                "defaultValue": True,
                "fullKey": "setting",
                "shortKey": "setting",
                "typeKey": "boolean",
                "typeParams": {
                    "displayName": "UI label",
                    "hint": "UI tooltip",
                },
            }
        ]
    }
    assert kv_config_schematics.to_dict() == expected_kv_config_schematics


def test_config_field_int() -> None:
    class ConfigSchema(BaseConfigSchema):
        setting: int = config_field(label="UI label", hint="UI tooltip", default=42)

    assert ConfigSchema.setting == 42
    assert ConfigSchema.setting == 42
    kv_config_schematics = ConfigSchema._to_kv_config_schematics()
    assert kv_config_schematics is not None
    expected_kv_config_schematics = {
        "fields": [
            {
                "defaultValue": 42,
                "fullKey": "setting",
                "shortKey": "setting",
                "typeKey": "numeric",
                "typeParams": {
                    "displayName": "UI label",
                    "hint": "UI tooltip",
                    "int": True,
                },
            }
        ]
    }
    assert kv_config_schematics.to_dict() == expected_kv_config_schematics


def test_config_field_float() -> None:
    class ConfigSchema(BaseConfigSchema):
        setting: float = config_field(label="UI label", hint="UI tooltip", default=4.2)

    assert ConfigSchema.setting == 4.2
    assert ConfigSchema().setting == 4.2
    kv_config_schematics = ConfigSchema._to_kv_config_schematics()
    assert kv_config_schematics is not None
    expected_kv_config_schematics = {
        "fields": [
            {
                "defaultValue": 4.2,
                "fullKey": "setting",
                "shortKey": "setting",
                "typeKey": "numeric",
                "typeParams": {
                    "displayName": "UI label",
                    "hint": "UI tooltip",
                    "int": False,
                },
            }
        ]
    }
    assert kv_config_schematics.to_dict() == expected_kv_config_schematics


def test_config_field_str() -> None:
    class ConfigSchema(BaseConfigSchema):
        setting: str = config_field(label="UI label", hint="UI tooltip", default="text")

    assert ConfigSchema.setting == "text"
    assert ConfigSchema().setting == "text"
    kv_config_schematics = ConfigSchema._to_kv_config_schematics()
    assert kv_config_schematics is not None
    expected_kv_config_schematics = {
        "fields": [
            {
                "defaultValue": "text",
                "fullKey": "setting",
                "shortKey": "setting",
                "typeKey": "string",
                "typeParams": {
                    "displayName": "UI label",
                    "hint": "UI tooltip",
                },
            }
        ]
    }
    assert kv_config_schematics.to_dict() == expected_kv_config_schematics



"""Allow execution of this subpackage as a script."""

import sys

from .cli import main

# Handle multiprocessing potentially re-running this module with a name other than `__main__`
if __name__ == "__main__":
    sys.exit(main())



#####################################################################
# This files has been automatically generated from:
#   ../async/test_repository_async.py
#
# DO NOT EDIT THIS FILE! Edit the async test case listed above,
# and regenerate the synchronous test cases with async2sync.py
#####################################################################
"""Test downloading models through the repository namespace."""

import logging

import pytest
from pytest import LogCaptureFixture as LogCap

from lmstudio import Client, LMStudioClientError

from ..support import SMALL_LLM_SEARCH_TERM


# N.B. We can maybe provide a reference list for what should be available
# if we narrow down the query enough, since it shouldn't really change.
# These also have to be tested in tandem because the model identifiers
# passed between methods are session-specific.
# You also need to delete the downloaded file after every test.
@pytest.mark.slow
@pytest.mark.lmstudio
def test_download_model_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        models = client.repository.search_models(SMALL_LLM_SEARCH_TERM)
        logging.info(f"Models: {models}")
        assert models
        assert isinstance(models, list)
        assert len(models) > 0

        options = models[0].get_download_options()
        logging.info(f"Download options: {options}")
        assert options
        assert isinstance(options, list)
        assert len(options) > 0

        model_path = options[0].download()
        logging.info(f"Downloaded model identifier: {model_path}")
        assert model_path
        assert isinstance(model_path, str)


@pytest.mark.slow
@pytest.mark.lmstudio
def test_get_options_out_of_session_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        models = client.repository.search_models(SMALL_LLM_SEARCH_TERM)
        assert models
        assert isinstance(models, list)
        assert len(models) > 0

    with pytest.raises(LMStudioClientError):
        models[0].get_download_options()


@pytest.mark.slow
@pytest.mark.lmstudio
def test_download_out_of_session_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        models = client.repository.search_models(SMALL_LLM_SEARCH_TERM)
        logging.info(f"Models: {models}")
        assert models
        assert isinstance(models, list)
        assert len(models) > 0

        options = models[0].get_download_options()
        assert options
        assert isinstance(options, list)
        assert len(options) > 0

    with pytest.raises(LMStudioClientError):
        options[0].download()



"""SDK logging support."""

import json
import logging

from typing import Any

# It's difficult to make a general purpose structured logging
# library play nice with arbitrary applications, so we instead
# implement rudimentary structured logging support based on the
# relevant standard library recipe:
#
# https://docs.python.org/3/howto/logging-cookbook.html#implementing-structured-logging

# TODO: Consider using context variables to pass event context state between API layers
#       In particular, associating json_api endpoint processing with the corresponding
#       multiplexing IDs in sync_api and async_api. Explicitly passing the IDs to the
#       message handling APIs purely for logging purposes would be outright annoying.

LogEventContext = dict[str, Any]


class StructuredLogEvent:
    def __init__(self, event: str, event_dict: LogEventContext) -> None:
        event_dict["event"] = event
        self.event = event
        self.event_dict = event_dict

    def as_formatted_json(self) -> str:
        return json.dumps(self.event_dict, sort_keys=True)

    def __str__(self) -> str:
        return str(self.as_formatted_json())


class StructuredLogger:
    def __init__(self, logger: logging.Logger) -> None:
        self._stdlib_logger = logger
        self.event_context: LogEventContext = {}

    def update_context(
        self, log_context: LogEventContext | None = None, /, **additional_context: Any
    ) -> None:
        event_context = self.event_context
        if log_context is not None:
            event_context.update(log_context)
        event_context.update(additional_context)

    def _log(
        self,
        level: int,
        msg: str,
        exc_info: bool,
        stack_info: bool,
        stacklevel: int,
        event_dict: LogEventContext,
    ) -> None:
        event_data = self.event_context.copy()
        event_data.update(event_dict)
        logged_msg = StructuredLogEvent(msg, event_data)
        stacklevel = stacklevel + 1
        self._stdlib_logger.log(
            level,
            logged_msg,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=event_data,
        )

    def log(
        self,
        level: int,
        msg: str,
        *,
        exc_info: bool = False,
        stack_info: bool = False,
        stacklevel: int = 1,
        **event_dict: Any,
    ) -> None:
        stacklevel = stacklevel + 1
        self._log(level, msg, exc_info, stack_info, stacklevel, event_dict)

    def debug(
        self,
        msg: str,
        *,
        exc_info: bool = False,
        stack_info: bool = False,
        stacklevel: int = 1,
        **event_dict: Any,
    ) -> None:
        stacklevel = stacklevel + 1
        self._log(logging.DEBUG, msg, exc_info, stack_info, stacklevel, event_dict)

    def info(
        self,
        msg: str,
        *,
        exc_info: bool = False,
        stack_info: bool = False,
        stacklevel: int = 1,
        **event_dict: Any,
    ) -> None:
        stacklevel = stacklevel + 1
        self._log(logging.INFO, msg, exc_info, stack_info, stacklevel, event_dict)

    def warn(
        self,
        msg: str,
        *,
        exc_info: bool = False,
        stack_info: bool = False,
        stacklevel: int = 1,
        **event_dict: Any,
    ) -> None:
        stacklevel = stacklevel + 1
        self._log(logging.WARN, msg, exc_info, stack_info, stacklevel, event_dict)

    def error(
        self,
        msg: str,
        *,
        exc_info: bool = False,
        stack_info: bool = False,
        stacklevel: int = 1,
        **event_dict: Any,
    ) -> None:
        stacklevel = stacklevel + 1
        self._log(logging.ERROR, msg, exc_info, stack_info, stacklevel, event_dict)

    def exception(self, msg: str, *, stacklevel: int = 1, **event_dict: Any) -> None:
        stacklevel = stacklevel + 1
        # Setting exc_info and stack_info is implied by calling this method instead of `error()`
        self._log(logging.ERROR, msg, True, True, stacklevel, event_dict)

    def critical(
        self,
        msg: str,
        *,
        exc_info: bool = False,
        stack_info: bool = False,
        stacklevel: int = 1,
        **event_dict: Any,
    ) -> None:
        stacklevel = stacklevel + 1
        self._log(logging.CRITICAL, msg, exc_info, stack_info, stacklevel, event_dict)


def new_logger(name: str, /, *args: Any, **kwds: Any) -> StructuredLogger:
    # Loggers contain runtime state, so they are NOT shared based on their names
    # Function name reflects this to avoid false expectations from stdlib logging
    return StructuredLogger(logging.getLogger(name, *args, **kwds))



"""Invoking and supporting plugin hook implementations."""
# Using wildcard imports to export API symbols is acceptable
# ruff: noqa: F403, F405

from .common import *
from .prompt_preprocessor import *
from .token_generator import *
from .tools_provider import *

# Available as lmstudio.plugin.*
__all__ = [
    "AsyncToolCallContext",
    "PromptPreprocessorController",
    "TokenGeneratorController",
    "ToolCallContext",
    "ToolsProviderController",
    "get_tool_call_context",
    "get_tool_call_context_async",
]



"""Test client timeout behaviour."""

import logging

from contextlib import contextmanager
from typing import Generator

import pytest
from pytest import LogCaptureFixture as LogCap

from lmstudio import (
    Client,
    LMStudioTimeoutError,
    get_sync_api_timeout,
    set_sync_api_timeout,
)
from lmstudio.sync_api import _DEFAULT_TIMEOUT

from .support import EXPECTED_LLM_ID

# Sync only, as async API uses standard async timeout constructs like anyio.move_on_after


@contextmanager
def sync_api_timeout(timeout: float | None) -> Generator[float | None, None, None]:
    previous_timeout = get_sync_api_timeout()
    set_sync_api_timeout(timeout)
    try:
        yield previous_timeout
    finally:
        set_sync_api_timeout(previous_timeout)


def test_default_timeout() -> None:
    # Ensure default timeout is defined, but is not excessively short or long
    # (the bounds that are considered reasonable may change over time)
    assert _DEFAULT_TIMEOUT is not None
    assert _DEFAULT_TIMEOUT >= 60
    assert _DEFAULT_TIMEOUT <= 600


@pytest.mark.parametrize("timeout", (None, 0, 1.5, 3600, 3600 * 24 * 7))
def test_timeout_updates_sync(timeout: float | None) -> None:
    with sync_api_timeout(timeout) as previous_timeout:
        assert previous_timeout == _DEFAULT_TIMEOUT
        assert get_sync_api_timeout() == timeout
    assert get_sync_api_timeout() == previous_timeout


@pytest.mark.lmstudio
def test_timeout_rpc_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)

    with Client() as client:
        model = client.llm.model(EXPECTED_LLM_ID)
        with sync_api_timeout(0):
            assert get_sync_api_timeout() == 0
            with pytest.raises(LMStudioTimeoutError):
                response = model.get_info()
                logging.error(f"Unexpected response: {response}")


@pytest.mark.lmstudio
def test_timeout_channel_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)

    with Client() as client:
        model = client.llm.model(EXPECTED_LLM_ID)
        with sync_api_timeout(0):
            assert get_sync_api_timeout() == 0
            with pytest.raises(LMStudioTimeoutError):
                response = model.respond("This will time out")
                logging.error(f"Unexpected response: {response}")



"""Generate sync API test cases from async API test cases."""
# Many of the async test cases can be transformed into valid
# sync API test cases via relatively naive text transforms.
# Test cases where that is the case are stored in the `async` folder.
# This script takes those test cases and emits sync API equivalents.

import re
import subprocess
from pathlib import Path
from os.path import commonpath

GENERATED_FILE_HEADER = """\
#####################################################################
# This files has been automatically generated from:
#   ../{}
#
# DO NOT EDIT THIS FILE! Edit the async test case listed above,
# and regenerate the synchronous test cases with async2sync.py
#####################################################################
"""


def convert_file(input_file: Path, sync_dir: Path) -> None:
    output_file = sync_dir / re.sub("_async", "_sync", input_file.name)
    print(f"Converting {input_file} to {output_file}")

    content = input_file.read_text()

    # Note: some replacements will only be relevant after dropping 3.10 support
    #       (as the native asyncio API were only added in 3.11)

    replacements = [
        (r"AsyncPredictionC", "SyncPredictionC"),  # Channel & CM
        (r"AsyncSession", "SyncSession"),
        (r"@pytest\.mark\.asyncio\n*", ""),
        (r"async def ([^(]*?)_async", r"def \1_sync"),
        (r"async def", "def"),
        (r"async for", "for"),
        (r"async with", "with"),
        (r"import asyncio\n", ""),
        (r"asyncio\.timeout\([^)]*\)", "nullcontext()"),
        (r"asyncio\.", ""),
        (r"aconnect_ws", "connect_ws"),
        (r"aclose", "close"),
        (r"aenter", "enter"),
        (r"aexit", "exit"),
        (r"anext", "next"),
        (r"Async", ""),
        (r"await +", ""),
        # Switch to native asyncio APIs when dropping Python 3.10 support
        (r"import anyio\n", ""),
        (r"anyio\.fail_after\([^)]*\)", "nullcontext()"),
    ]

    for pattern in replacements:
        content = re.sub(pattern[0], pattern[1], content)

    # need to import from contextlib if we use nullcontext
    if "nullcontext()" in content:
        # add it after the logging import, this is always present
        content = re.sub(
            r"import logging\n",
            "import logging\nfrom contextlib import nullcontext\n",
            content,
        )

    # Use commonpath, as there's no `walk_up` parameter on `relative_to`` in Python 3.11
    base_path = commonpath((input_file, output_file))
    relative_input_path = str(input_file.relative_to(base_path).as_posix())
    output_header = GENERATED_FILE_HEADER.format(relative_input_path)
    output_text = output_header + content
    output_file.write_text(output_text)


def main() -> None:
    async_dir = Path(__file__).parent / "async"
    sync_dir = Path(__file__).parent / "sync"
    async_files = async_dir.glob("test_*_async.py")

    for input_file in async_files:
        convert_file(input_file, sync_dir)
    print("Running automatic formatter after file conversion")
    subprocess.run(["tox", "-e", "format"])


if __name__ == "__main__":
    main()



"""Emulated "SDK" for the traceback filtering unit tests."""

from pathlib import Path
from typing import AsyncGenerator, Callable, Coroutine, Generator, NoReturn, TypeAlias

from lmstudio import LMStudioError
from lmstudio.sdk_api import sdk_public_api, sdk_public_api_async, sdk_public_type

# The traceback filtering tests rely on this file being in
# an "lmstudio" folder (similar to the real SDK)
assert __name__.rpartition(".")[2] == "lmstudio"
assert "lmstudio" in Path(__file__).parts


def raise_sdk_error() -> NoReturn:
    raise LMStudioError("Emulate anticipated runtime error")


def raise_external_error() -> NoReturn:
    raise BaseException("Emulate Ctrl-C")


def raise_internal_error() -> NoReturn:
    raise Exception("Emulate unexpected runtime error")


ErrFunc: TypeAlias = Callable[[], NoReturn]


def internal_func(err_func: ErrFunc) -> NoReturn:
    err_func()


@sdk_public_api()
def public_func(err_func: ErrFunc) -> NoReturn:
    internal_func(err_func)


@sdk_public_api()
def public_wrapper_func(err_func: ErrFunc) -> NoReturn:
    public_func(err_func)


class InternalIterator:
    # Define an iterable class to align with the way
    # AsyncPrediction and Prediction are implemented
    def __init__(self, err_func: ErrFunc) -> None:
        self._err_func = err_func


class AsyncInternalIterator(InternalIterator):
    async def __aiter__(self) -> AsyncGenerator[None, NoReturn]:
        yield None
        internal_func(self._err_func)


class SyncInternalIterator(InternalIterator):
    def __iter__(self) -> Generator[None, None, NoReturn]:
        yield None
        internal_func(self._err_func)


@sdk_public_api()
def public_iter_wrapper(err_func: ErrFunc) -> None:
    for _ in SyncInternalIterator(err_func):
        pass


@sdk_public_api_async()
async def public_coroutine(err_func: ErrFunc) -> NoReturn:
    internal_func(err_func)


@sdk_public_api_async()
async def public_wrapper_coroutine(err_func: ErrFunc) -> NoReturn:
    await public_coroutine(err_func)


@sdk_public_api_async()
async def public_asynciter_wrapper(err_func: ErrFunc) -> None:
    async for _ in AsyncInternalIterator(err_func):
        pass


@sdk_public_type
class PublicClass:
    def internal_method(self, err_func: ErrFunc) -> NoReturn:
        internal_func(err_func)

    @sdk_public_api()
    def public_method(self, err_func: ErrFunc) -> NoReturn:
        self.internal_method(err_func)

    @sdk_public_api()
    def public_wrapper_method(self, err_func: ErrFunc) -> NoReturn:
        self.public_method(err_func)

    @sdk_public_api()
    def public_iter_wrapper_method(self, err_func: ErrFunc) -> None:
        for _ in SyncInternalIterator(err_func):
            pass

    @sdk_public_api_async()
    async def public_async_method(self, err_func: ErrFunc) -> NoReturn:
        internal_func(err_func)

    @sdk_public_api_async()
    async def public_async_wrapper_method(self, err_func: ErrFunc) -> NoReturn:
        await public_coroutine(err_func)

    @sdk_public_api_async()
    async def public_asynciter_wrapper_method(self, err_func: ErrFunc) -> None:
        async for _ in AsyncInternalIterator(err_func):
            pass


TestFunc: TypeAlias = Callable[[ErrFunc], NoReturn | None]
TestCoro: TypeAlias = Callable[[ErrFunc], Coroutine[None, None, NoReturn | None]]

_PC = PublicClass()

SYNC_API: tuple[TestFunc, ...] = (
    public_func,
    public_wrapper_func,
    public_iter_wrapper,
    _PC.public_method,
    _PC.public_wrapper_method,
    _PC.public_iter_wrapper_method,
)
ASYNC_API: tuple[TestCoro, ...] = (
    public_coroutine,
    public_wrapper_coroutine,
    public_asynciter_wrapper,
    _PC.public_async_method,
    _PC.public_async_wrapper_method,
    _PC.public_asynciter_wrapper_method,
)



"""Test chat history management."""

import copy
import json

from typing import Callable, cast

import pytest

from lmstudio.sdk_api import LMStudioOSError
from lmstudio.schemas import DictObject
from lmstudio.history import (
    AnyChatMessageDict,
    AnyChatMessageInput,
    AssistantMultiPartInput,
    Chat,
    ChatHistoryData,
    ChatHistoryDataDict,
    LocalFileInput,
    FileHandle,
    _FileHandleCache,
    FileHandleDict,
    _LocalFileData,
    TextData,
)
from lmstudio.json_api import (
    LlmInfo,
    LlmLoadModelConfig,
    LlmPredictionConfig,
    LlmPredictionStats,
    PredictionResult,
)
from lmstudio._sdk_models import (
    ToolCallRequestDataDict,
    ToolCallResultDataDict,
)

from .support import IMAGE_FILEPATH, check_sdk_error

INPUT_ENTRIES: list[DictObject] = [
    # Entries with multi-word keys mix snake_case and camelCase
    # to ensure both are accepted and normalized (to camelCase)
    {
        "role": "system",
        "content": [
            {"type": "text", "text": "Initial system messages"},
        ],
    },
    {
        "role": "user",
        "content": [{"type": "text", "text": "Simple text prompt"}],
    },
    {
        "role": "user",
        "content": [{"type": "text", "text": "Structured text prompt"}],
    },
    {
        "role": "user",
        "content": [
            {
                "type": "file",
                "name": "someFile.txt",
                "identifier": "some-file",
                "size_bytes": 100,
                "fileType": "text/plain",
            }
        ],
    },
    {
        "role": "user",
        "content": [
            {
                "type": "file",
                "name": "someOtherFile.txt",
                "identifier": "some-other-file",
                "sizeBytes": 100,
                "file_type": "text/plain",
            }
        ],
    },
    {
        "role": "system",
        "content": [{"type": "text", "text": "Simple text system prompt"}],
    },
    {
        "role": "assistant",
        "content": [{"type": "text", "text": "Simple text response"}],
    },
    {
        "role": "user",
        "content": [{"type": "text", "text": "Avoid consecutive responses"}],
    },
    {
        "role": "assistant",
        "content": [{"type": "text", "text": "Structured text response"}],
    },
    {
        "role": "user",
        "content": [{"type": "text", "text": "Avoid consecutive responses"}],
    },
    {
        "role": "assistant",
        "content": [
            {
                "type": "file",
                "name": "someFile.txt",
                "identifier": "some-file",
                "size_bytes": 100,
                "fileType": "text/plain",
            }
        ],
    },
    {
        "role": "user",
        "content": [{"type": "text", "text": "Avoid consecutive responses"}],
    },
    {
        "role": "assistant",
        "content": [
            {
                "type": "file",
                "name": "someOtherFile.txt",
                "identifier": "some-other-file",
                "sizeBytes": 100,
                "file_type": "text/plain",
            }
        ],
    },
    {
        "role": "system",
        "content": [{"type": "text", "text": "Structured text system prompt"}],
    },
    {
        "role": "assistant",
        "content": [
            {"type": "text", "text": "Example tool call request"},
            {
                "type": "toolCallRequest",
                "toolCallRequest": {
                    "type": "function",
                    "id": "114663647",
                    "name": "example_tool_name",
                    "arguments": {
                        "n": 58013,
                        "t": "value",
                    },
                },
            },
            {
                "type": "toolCallRequest",
                "toolCallRequest": {
                    "type": "function",
                    "id": "114663648",
                    "name": "another_example_tool_name",
                    "arguments": {
                        "n": 23,
                        "t": "some other value",
                    },
                },
            },
        ],
    },
    {
        "role": "tool",
        "content": [
            {
                "type": "toolCallResult",
                "toolCallId": "114663647",
                "content": "example tool call result",
            },
            {
                "type": "toolCallResult",
                "toolCallId": "114663648",
                "content": "another example tool call result",
            },
        ],
    },
]

INPUT_HISTORY = {"messages": INPUT_ENTRIES}

# Consecutive user and assistant messages should be merged
EXPECTED_MESSAGES: list[AnyChatMessageDict] = [
    {
        "role": "system",
        "content": [
            {"type": "text", "text": "Initial system messages"},
        ],
    },
    {
        "role": "user",
        "content": [
            {"type": "text", "text": "Simple text prompt"},
            {"type": "text", "text": "Structured text prompt"},
            {
                "type": "file",
                "name": "someFile.txt",
                "identifier": "some-file",
                "sizeBytes": 100,
                "fileType": "text/plain",
            },
            {
                "type": "file",
                "name": "someOtherFile.txt",
                "identifier": "some-other-file",
                "sizeBytes": 100,
                "fileType": "text/plain",
            },
        ],
    },
    {
        "role": "system",
        "content": [{"type": "text", "text": "Simple text system prompt"}],
    },
    {
        "role": "assistant",
        "content": [
            {"type": "text", "text": "Simple text response"},
        ],
    },
    {
        "role": "user",
        "content": [{"type": "text", "text": "Avoid consecutive responses"}],
    },
    {
        "role": "assistant",
        "content": [
            {"type": "text", "text": "Structured text response"},
        ],
    },
    {
        "role": "user",
        "content": [{"type": "text", "text": "Avoid consecutive responses"}],
    },
    {
        "role": "assistant",
        "content": [
            {
                "type": "file",
                "name": "someFile.txt",
                "identifier": "some-file",
                "sizeBytes": 100,
                "fileType": "text/plain",
            },
        ],
    },
    {
        "role": "user",
        "content": [{"type": "text", "text": "Avoid consecutive responses"}],
    },
    {
        "role": "assistant",
        "content": [
            {
                "type": "file",
                "name": "someOtherFile.txt",
                "identifier": "some-other-file",
                "sizeBytes": 100,
                "fileType": "text/plain",
            },
        ],
    },
    {
        "role": "system",
        "content": [{"type": "text", "text": "Structured text system prompt"}],
    },
    {
        "role": "assistant",
        "content": [
            {"type": "text", "text": "Example tool call request"},
            {
                "type": "toolCallRequest",
                "toolCallRequest": {
                    "type": "function",
                    "id": "114663647",
                    "name": "example_tool_name",
                    "arguments": {
                        "n": 58013,
                        "t": "value",
                    },
                },
            },
            {
                "type": "toolCallRequest",
                "toolCallRequest": {
                    "type": "function",
                    "id": "114663648",
                    "name": "another_example_tool_name",
                    "arguments": {
                        "n": 23,
                        "t": "some other value",
                    },
                },
            },
        ],
    },
    {
        "role": "tool",
        "content": [
            {
                "type": "toolCallResult",
                "toolCallId": "114663647",
                "content": "example tool call result",
            },
            {
                "type": "toolCallResult",
                "toolCallId": "114663648",
                "content": "another example tool call result",
            },
        ],
    },
]


EXPECTED_HISTORY: ChatHistoryDataDict = {"messages": EXPECTED_MESSAGES}


def test_from_history() -> None:
    # We *expect* the input to fail static typechecking here,
    # as it's relying on the convenience input transformations
    chat = Chat.from_history(INPUT_HISTORY)  # type: ignore[arg-type]
    assert chat._get_history_for_prediction() == EXPECTED_HISTORY
    cloned_chat = Chat.from_history(chat._history)
    assert cloned_chat._get_history_for_prediction() == EXPECTED_HISTORY


def test_from_history_with_simple_text() -> None:
    message = "This is a basic text message"
    # Plain string should be converted to a single user message
    expected_text_content = {"type": "text", "text": message}
    expected_user_message = {
        "role": "user",
        "content": [expected_text_content],
    }
    expected_history = {"messages": [expected_user_message]}
    chat = Chat.from_history(message)
    assert chat._get_history_for_prediction() == expected_history
    # Plain strings should also be accepted as a text content field
    input_history = {
        "messages": [
            {"role": "user", "content": message},
            {"role": "system", "content": message},
        ]
    }
    expected_system_message = {
        "role": "system",
        "content": [expected_text_content],
    }
    expected_history = {"messages": [expected_user_message, expected_system_message]}
    # We *expect* the input to fail static typechecking here,
    # as it's relying on the convenience input transformations
    chat = Chat.from_history(input_history)  # type: ignore[arg-type]
    assert chat._get_history_for_prediction() == expected_history


INPUT_FILE_HANDLE = FileHandle(
    name="someFile.txt",
    identifier="some-file",
    size_bytes=100,
    file_type="text/plain",
)
INPUT_FILE_HANDLE_DICT: FileHandleDict = {
    "type": "file",
    "name": "someOtherFile.txt",
    "identifier": "some-other-file",
    "sizeBytes": 100,
    "fileType": "text/plain",
}
INPUT_TOOL_REQUESTS: list[ToolCallRequestDataDict] = [
    {
        "type": "toolCallRequest",
        "toolCallRequest": {
            "type": "function",
            "id": "114663647",
            "name": "example_tool_name",
            "arguments": {
                "n": 58013,
                "t": "value",
            },
        },
    },
    {
        "type": "toolCallRequest",
        "toolCallRequest": {
            "type": "function",
            "id": "114663648",
            "name": "another_example_tool_name",
            "arguments": {
                "n": 23,
                "t": "some other value",
            },
        },
    },
]
INPUT_TOOL_RESULTS: list[ToolCallResultDataDict] = [
    {
        "type": "toolCallResult",
        "toolCallId": "114663647",
        "content": "example tool call result",
    },
    {
        "type": "toolCallResult",
        "toolCallId": "114663648",
        "content": "another example tool call result",
    },
]


def test_get_history() -> None:
    # Also tests the specific message addition methods
    chat = Chat("Initial system messages")
    chat.add_user_message("Simple text prompt")
    chat.add_user_message(TextData(text="Structured text prompt"))
    chat.add_user_message(INPUT_FILE_HANDLE)
    chat.add_user_message(INPUT_FILE_HANDLE_DICT)
    chat.add_system_prompt("Simple text system prompt")
    chat.add_assistant_response("Simple text response")
    chat.add_user_message("Avoid consecutive responses")
    chat.add_assistant_response(TextData(text="Structured text response"))
    chat.add_user_message("Avoid consecutive responses")
    chat.add_assistant_response(INPUT_FILE_HANDLE)
    chat.add_user_message("Avoid consecutive responses")
    chat.add_assistant_response(INPUT_FILE_HANDLE_DICT)
    chat.add_system_prompt(TextData(text="Structured text system prompt"))
    chat.add_assistant_response("Example tool call request", INPUT_TOOL_REQUESTS)
    chat.add_tool_results(INPUT_TOOL_RESULTS)
    assert chat._get_history_for_prediction() == EXPECTED_HISTORY


def test_add_entry() -> None:
    chat = Chat("Initial system messages")
    chat.add_entry("user", "Simple text prompt")
    chat.add_entry("user", TextData(text="Structured text prompt"))
    chat.add_entry("user", INPUT_FILE_HANDLE)
    chat.add_entry("user", INPUT_FILE_HANDLE_DICT)
    chat.add_entry("system", "Simple text system prompt")
    chat.add_entry("assistant", "Simple text response")
    chat.add_entry("user", "Avoid consecutive responses")
    chat.add_entry("assistant", TextData(text="Structured text response"))
    chat.add_entry("user", "Avoid consecutive responses")
    chat.add_entry("assistant", INPUT_FILE_HANDLE)
    chat.add_entry("user", "Avoid consecutive responses")
    chat.add_entry("assistant", INPUT_FILE_HANDLE_DICT)
    chat.add_entry("system", TextData(text="Structured text system prompt"))
    tool_call_message_contents: AssistantMultiPartInput = [
        "Example tool call request",
        *INPUT_TOOL_REQUESTS,
    ]
    chat.add_entry("assistant", tool_call_message_contents)
    chat.add_entry("tool", INPUT_TOOL_RESULTS)
    assert chat._get_history_for_prediction() == EXPECTED_HISTORY


def test_append() -> None:
    chat = Chat()
    for message in INPUT_ENTRIES:
        chat.append(cast(AnyChatMessageDict, message))
    assert chat._get_history_for_prediction() == EXPECTED_HISTORY


def test_add_entries_dict_content() -> None:
    history_data = EXPECTED_MESSAGES
    chat = Chat()
    chat._add_entries(history_data)
    assert chat._get_history_for_prediction() == EXPECTED_HISTORY


def test_add_entries_tuple_content() -> None:
    history_data: list[tuple[str, AnyChatMessageInput]] = [
        (m["role"], cast(AnyChatMessageInput, m["content"])) for m in EXPECTED_MESSAGES
    ]
    chat = Chat()
    chat._add_entries(history_data)
    assert chat._get_history_for_prediction() == EXPECTED_HISTORY


def test_add_entries_class_content() -> None:
    history_data = ChatHistoryData.from_dict(EXPECTED_HISTORY).messages
    chat = Chat()
    chat._add_entries(history_data)
    assert chat._get_history_for_prediction() == EXPECTED_HISTORY


def _make_prediction_result(data: str | DictObject) -> PredictionResult:
    return PredictionResult(
        content=(data if isinstance(data, str) else json.dumps(data)),
        parsed=data,
        stats=LlmPredictionStats(stop_reason="failed"),
        model_info=LlmInfo(
            model_key="model-id",
            path="model/path",
            format="gguf",
            display_name="Some LLM",
            size_bytes=0,
            vision=False,
            trained_for_tool_use=False,
            max_context_length=32,
        ),
        load_config=LlmLoadModelConfig(),
        prediction_config=LlmPredictionConfig(),
    )


EXPECTED_PREDICTION_RESPONSE_HISTORY = {
    "messages": [
        {
            "role": "assistant",
            "content": [{"type": "text", "text": "Unstructured prediction"}],
        },
        {
            "role": "user",
            "content": [{"type": "text", "text": "Avoid consecutive responses."}],
        },
        {
            "role": "assistant",
            "content": [{"type": "text", "text": '{"structured": "prediction"}'}],
        },
    ]
}


def test_add_prediction_results() -> None:
    chat = Chat()
    chat.add_assistant_response(_make_prediction_result("Unstructured prediction"))
    chat.add_user_message("Avoid consecutive responses.")
    chat.add_assistant_response(_make_prediction_result({"structured": "prediction"}))
    # Note: file handles are not yet supported in prediction responses
    assert chat._get_history_for_prediction() == EXPECTED_PREDICTION_RESPONSE_HISTORY


EXPECTED_PENDING_FILE_HANDLES = [
    {
        "fileType": "unknown",
        "identifier": "<file addition pending>",
        "name": "raw-binary.txt",
        "sizeBytes": -1,
        "type": "file",
    },
    {
        "fileType": "unknown",
        "identifier": "<file addition pending>",
        "name": "raw-binary.txt",
        "sizeBytes": -1,
        "type": "file",
    },
    {
        "fileType": "unknown",
        "identifier": "<file addition pending>",
        "name": "lemmy.png",
        "sizeBytes": -1,
        "type": "file",
    },
    {
        "fileType": "unknown",
        "identifier": "<file addition pending>",
        "name": "also-lemmy.png",
        "sizeBytes": -1,
        "type": "file",
    },
    {
        "fileType": "unknown",
        "identifier": "<file addition pending>",
        "name": "lemmy.png",
        "sizeBytes": -1,
        "type": "file",
    },
]


EXPECTED_RESOLVED_FILE_HANDLES: list[DictObject] = [
    {
        "fileType": "text/plain",
        "identifier": "file-1",
        "name": "raw-binary.txt",
        "sizeBytes": 20,
        "type": "file",
    },
    {
        "fileType": "text/plain",
        "identifier": "file-1",
        "name": "raw-binary.txt",
        "sizeBytes": 20,
        "type": "file",
    },
    {
        "fileType": "image",
        "identifier": "file-2",
        "name": "lemmy.png",
        "sizeBytes": 41812,
        "type": "file",
    },
    {
        "fileType": "image",
        "identifier": "file-3",
        "name": "also-lemmy.png",
        "sizeBytes": 41812,
        "type": "file",
    },
    {
        "fileType": "image",
        "identifier": "file-2",
        "name": "lemmy.png",
        "sizeBytes": 41812,
        "type": "file",
    },
]


def _add_file(file_data: _LocalFileData, identifier: str) -> FileHandle:
    name = file_data.name
    fetch_param = file_data._as_fetch_param()
    return FileHandle(
        name=name,
        identifier=identifier,
        size_bytes=len(fetch_param.content_base64),
        file_type="image" if name.endswith(".png") else "text/plain",
    )


def _check_pending_file(file_handle: FileHandle, name: str) -> None:
    assert file_handle.type == "file"
    assert file_handle.name == name
    assert file_handle.identifier == "<file addition pending>"
    assert file_handle.size_bytes == -1
    assert file_handle.file_type == "unknown"


def _check_fetched_text_file(
    file_handle: FileHandle, name: str, identifier: str
) -> None:
    assert file_handle.type == "file"
    assert file_handle.name == name
    assert file_handle.identifier == identifier
    assert file_handle.size_bytes > 0
    assert file_handle.file_type == "text/plain"


def _make_local_file_cache() -> tuple[_FileHandleCache, list[FileHandle], int]:
    # File context for fetching handles that ensures
    # * duplicate files are only looked up once
    # * files with different names are looked up under both names
    cache = _FileHandleCache()
    num_unique_files = 3
    files_to_cache: list[tuple[LocalFileInput, str | None]] = [
        (b"raw binary data", "raw-binary.txt"),
        (b"raw binary data", "raw-binary.txt"),
        (IMAGE_FILEPATH, None),
        (IMAGE_FILEPATH, "also-lemmy.png"),
        (IMAGE_FILEPATH, None),
    ]
    file_handles: list[FileHandle] = []
    for args in files_to_cache:
        file_handles.append(cache._get_file_handle(*args))
    assert [h.to_dict() for h in file_handles] == EXPECTED_PENDING_FILE_HANDLES
    return cache, file_handles, num_unique_files


# TODO: Improve code sharing between this test case and its async counterpart
#       (potentially by moving the async version to `async/test_history_async.py`)
def test_file_handle_cache() -> None:
    local_files: list[_LocalFileData] = []
    unique_file_handles: list[FileHandle] = []

    def add_file(file_data: _LocalFileData) -> FileHandle:
        local_files.append(file_data)
        result = _add_file(file_data, f"file-{len(local_files)}")
        unique_file_handles.append(result)
        return result

    cache, file_handles, num_unique_files = _make_local_file_cache()
    cache._fetch_file_handles(add_file)
    assert len(local_files) == num_unique_files
    assert len(unique_file_handles) == num_unique_files
    assert [h.to_dict() for h in file_handles] == EXPECTED_RESOLVED_FILE_HANDLES
    # Adding the same file again should immediately populate the handle
    image_handle = cache._get_file_handle(IMAGE_FILEPATH)
    assert image_handle == file_handles[-1]
    # Fetching again should not perform any lookups
    cache._fetch_file_handles(add_file)
    assert len(local_files) == num_unique_files
    assert len(unique_file_handles) == num_unique_files
    # Adding a different file should require a new lookup
    this_file_handle = cache._get_file_handle(__file__)
    expected_name = f"{__name__.rpartition('.')[2]}.py"
    _check_pending_file(this_file_handle, expected_name)
    cache._fetch_file_handles(add_file)
    assert len(local_files) == num_unique_files + 1
    assert len(unique_file_handles) == num_unique_files + 1
    expected_identifier = f"file-{num_unique_files + 1}"
    _check_fetched_text_file(this_file_handle, expected_name, expected_identifier)


@pytest.mark.asyncio
async def test_file_handle_cache_async() -> None:
    local_files: list[_LocalFileData] = []
    unique_file_handles: list[FileHandle] = []

    async def add_file(file_data: _LocalFileData) -> FileHandle:
        local_files.append(file_data)
        result = _add_file(file_data, f"file-{len(local_files)}")
        unique_file_handles.append(result)
        return result

    cache, file_handles, num_unique_files = _make_local_file_cache()
    await cache._fetch_file_handles_async(add_file)
    assert len(local_files) == num_unique_files
    assert len(unique_file_handles) == num_unique_files
    assert [h.to_dict() for h in file_handles] == EXPECTED_RESOLVED_FILE_HANDLES
    # Adding the same file again should immediately populate the handle
    image_handle = cache._get_file_handle(IMAGE_FILEPATH)
    assert image_handle == file_handles[-1]
    # Fetching again should not perform any lookups
    await cache._fetch_file_handles_async(add_file)
    assert len(local_files) == num_unique_files
    assert len(unique_file_handles) == num_unique_files
    # Adding a different file should require a new lookup
    this_file_handle = cache._get_file_handle(__file__)
    expected_name = f"{__name__.rpartition('.')[2]}.py"
    _check_pending_file(this_file_handle, expected_name)
    await cache._fetch_file_handles_async(add_file)
    assert len(local_files) == num_unique_files + 1
    assert len(unique_file_handles) == num_unique_files + 1
    expected_identifier = f"file-{num_unique_files + 1}"
    _check_fetched_text_file(this_file_handle, expected_name, expected_identifier)


def test_invalid_local_file() -> None:
    cache = _FileHandleCache()
    with pytest.raises(LMStudioOSError) as exc_info:
        cache._get_file_handle("No such file")
    check_sdk_error(exc_info, __file__)


EXPECTED_USER_ATTACHMENT_MESSAGES = [
    {
        "content": [
            {
                "text": "What do you make of this?",
                "type": "text",
            },
            # Implementation attaches the prepared file handles
            # before it attaches the prepared image handles
            {
                "fileType": "text/plain",
                "identifier": "some-file",
                "name": "someFile.txt",
                "sizeBytes": 100,
                "type": "file",
            },
            {
                "fileType": "image",
                "identifier": "some-image",
                "name": "lemmy.png",
                "sizeBytes": 41812,
                "type": "file",
            },
        ],
        "role": "user",
    },
]

INPUT_IMAGE_HANDLE = FileHandle(
    name="lemmy.png",
    identifier="some-image",
    size_bytes=41812,
    file_type="image",
)


def test_user_message_attachments() -> None:
    chat = Chat()
    chat.add_user_message(
        "What do you make of this?",
        images=[INPUT_IMAGE_HANDLE],
        _files=[INPUT_FILE_HANDLE],
    )
    history = chat._get_history()
    assert history["messages"] == EXPECTED_USER_ATTACHMENT_MESSAGES


def test_assistant_responses_cannot_be_multipart_or_consecutive() -> None:
    chat = Chat()
    chat.add_assistant_response("First response")
    with pytest.raises(RuntimeError, match="Multi-part or consecutive"):
        chat.add_assistant_response("Consecutive response")
    chat.add_user_message("Separator")
    with pytest.raises(ValueError, match="Unable to parse"):
        # MyPy picks up that this is not allowed, but we want to test it anyway
        chat.add_assistant_response(("Multi-part", "response"))  # type: ignore[arg-type]
    chat.add_assistant_response("Second response")


def test_system_prompts_cannot_be_multipart_or_consecutive() -> None:
    chat = Chat("First prompt")
    with pytest.raises(RuntimeError, match="Multi-part or consecutive"):
        chat.add_system_prompt("Consecutive prompt")
    chat.add_user_message("Separator")
    with pytest.raises(ValueError, match="Unable to parse"):
        # MyPy picks up that this is not allowed, but we want to test it anyway
        chat.add_system_prompt(("Multi-part", "prompt"))  # type: ignore[arg-type]
    chat.add_system_prompt("Second prompt")


def test_system_prompts_cannot_be_file_handles() -> None:
    chat = Chat()
    with pytest.raises(ValueError, match="Unable to parse system prompt"):
        # MyPy picks up that this is not allowed, but we want to test it anyway
        chat.add_system_prompt(INPUT_FILE_HANDLE)  # type: ignore[arg-type]
    with pytest.raises(ValueError, match="Unable to parse system prompt"):
        chat.add_entry("system", INPUT_FILE_HANDLE)
    with pytest.raises(ValueError, match="Unable to parse system prompt"):
        # MyPy picks up that this is not allowed, but we want to test it anyway
        chat.add_system_prompt(INPUT_FILE_HANDLE_DICT)  # type: ignore[arg-type]
    with pytest.raises(ValueError, match="Unable to parse system prompt"):
        chat.add_entry("system", INPUT_FILE_HANDLE_DICT)


def test_initial_history_with_prompt_is_disallowed() -> None:
    chat = Chat()
    with pytest.raises(ValueError, match="initial history or a system prompt"):
        Chat("Initial prompt", _initial_history=chat._history)


EXPECTED_CHAT_STR = """\
Chat.from_history({
  "messages": [
    {
      "content": [
        {
          "text": "Initial system prompt",
          "type": "text"
        }
      ],
      "role": "system"
    },
    {
      "content": [
        {
          "text": "Simple text message",
          "type": "text"
        }
      ],
      "role": "user"
    }
  ]
})\
"""


def test_chat_display() -> None:
    chat = Chat("Initial system prompt")
    chat.add_user_message("Simple text message")
    # Chats use the standard identity based repr
    assert repr(chat) == object.__repr__(chat)
    # But print the history
    print(chat)
    assert str(chat) == EXPECTED_CHAT_STR


CLONING_MECHANISMS: list[Callable[[Chat], Chat]] = [
    Chat.from_history,
    Chat.copy,
    copy.copy,
    copy.deepcopy,
]


@pytest.mark.parametrize("clone", CLONING_MECHANISMS)
def test_chat_duplication(clone: Callable[[Chat], Chat]) -> None:
    chat = Chat("Initial system prompt")
    chat.add_user_message("Simple text message")
    cloned_chat = clone(chat)
    assert cloned_chat is not chat
    for attr, source_value in chat.__dict__.items():
        assert getattr(cloned_chat, attr) is not source_value
    chat_messages = chat._history.messages
    cloned_messages = cloned_chat._history.messages
    for source_message, cloned_message in zip(chat_messages, cloned_messages):
        assert cloned_message is not source_message
        assert cloned_message == source_message



"""Ensure the traceback filtering works as expected."""

import pytest
from pytest import LogCaptureFixture as LogCap

from lmstudio import LMStudioError
from lmstudio.sdk_api import sdk_callback_invocation
from lmstudio._logging import new_logger

from .support import check_sdk_error, check_unfiltered_error
from .support.lmstudio import (
    TestCoro,
    TestFunc,
    SYNC_API,
    ASYNC_API,
    raise_external_error,
    raise_internal_error,
    raise_sdk_error,
)


@pytest.mark.parametrize("public_api", SYNC_API)
def test_sync_api_truncation_sdk_error(public_api: TestFunc) -> None:
    with pytest.raises(LMStudioError) as exc_info:
        public_api(raise_sdk_error)
    check_sdk_error(exc_info, __file__)


@pytest.mark.parametrize("public_api", SYNC_API)
def test_sync_api_truncation_external_error(public_api: TestFunc) -> None:
    with pytest.raises(BaseException) as exc_info:
        public_api(raise_external_error)
    check_sdk_error(exc_info, __file__, check_exc=False)


@pytest.mark.parametrize("public_api", SYNC_API)
def test_sync_api_truncation_internal_error(public_api: TestFunc) -> None:
    with pytest.raises(Exception) as exc_info:
        public_api(raise_internal_error)
    check_unfiltered_error(exc_info, __file__, raise_internal_error)


@pytest.mark.asyncio
@pytest.mark.parametrize("public_api", ASYNC_API)
async def test_async_api_truncation_sdk_error(public_api: TestCoro) -> None:
    with pytest.raises(LMStudioError) as exc_info:
        await public_api(raise_sdk_error)
    check_sdk_error(exc_info, __file__)


@pytest.mark.asyncio
@pytest.mark.parametrize("public_api", ASYNC_API)
async def test_async_api_truncation_external_error(public_api: TestCoro) -> None:
    with pytest.raises(BaseException) as exc_info:
        await public_api(raise_external_error)
    check_sdk_error(exc_info, __file__, check_exc=False)


@pytest.mark.asyncio
@pytest.mark.parametrize("public_api", ASYNC_API)
async def test_async_api_truncation_internal_error(public_api: TestCoro) -> None:
    with pytest.raises(Exception) as exc_info:
        await public_api(raise_internal_error)
    check_unfiltered_error(exc_info, __file__, raise_internal_error)


def test_callback_invocation(caplog: LogCap) -> None:
    logger = new_logger(__name__)
    exc_to_raise = Exception("This will be raised")
    with sdk_callback_invocation("Callback test", logger):
        raise exc_to_raise
    logged_errors = caplog.get_records("call")
    logged_error = logged_errors[0]
    logged_msg = logged_error.getMessage()
    assert "Callback test" in logged_msg
    assert "This will be raised" in logged_msg
    assert logged_error.levelname.lower() == "error"
    assert len(logged_errors) == 1



"""Test common client session behaviour."""

import logging
from typing import Generator

import pytest
from pytest import LogCaptureFixture as LogCap

from lmstudio import (
    AsyncClient,
    Client,
    LMStudioWebsocketError,
)
from lmstudio.async_api import (
    _AsyncLMStudioWebsocket,
    _AsyncSession,
    _AsyncSessionSystem,
)
from lmstudio.sync_api import (
    SyncLMStudioWebsocket,
    _SyncSession,
    _SyncSessionSystem,
)
from lmstudio._ws_impl import AsyncTaskManager
from lmstudio._ws_thread import AsyncWebsocketThread


async def check_connected_async_session(session: _AsyncSession) -> None:
    assert session.connected
    session_ws = session._lmsws
    assert session_ws is not None
    assert session_ws.connected
    # Attempting explicit reconnection fails when connected
    with pytest.raises(LMStudioWebsocketError, match="already connected"):
        await session.connect()
    # Reentering a session has no effect if the websocket is open
    async with session:
        assert session.connected
        assert session_ws.connected
        assert session._lmsws is session_ws
    # But the session is closed after the *first* CM exit
    assert not session.connected
    assert not session_ws.connected


@pytest.mark.asyncio
@pytest.mark.lmstudio
async def test_session_cm_async(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    api_host = await AsyncClient.find_default_local_api_host()
    client = AsyncClient(api_host)
    session = _AsyncSessionSystem(client)
    # Sessions start out disconnected
    assert not session.connected
    # Disconnecting should run without error
    await session.disconnect()
    # Entering a session opens the websocket if it isn't already open
    async with client._task_manager, session as entry_result:
        # Sessions are their own entry result
        assert entry_result is session
        # Check connected session behaves as expected
        await check_connected_async_session(session)


# Check the synchronous session API


def check_connected_sync_session(session: _SyncSession) -> None:
    assert session.connected
    session_ws = session._lmsws
    assert session_ws is not None
    assert session_ws.connected
    # Attempting explicit reconnection fails when connected
    with pytest.raises(LMStudioWebsocketError, match="already connected"):
        session.connect()
    # Reentering a session has no effect if the websocket is open
    with session:
        assert session.connected
        assert session_ws.connected
        assert session._lmsws is session_ws
    # But the session is closed after the *first* CM exit
    assert not session.connected
    assert not session_ws.connected


@pytest.mark.lmstudio
def test_session_cm_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    api_host = Client.find_default_local_api_host()
    client = Client(api_host)
    session = _SyncSessionSystem(client)
    # Sessions start out disconnected
    assert not session.connected
    # Disconnecting should run without error
    session.disconnect()
    # Entering a session opens the websocket if it isn't already open
    with session as entry_result:
        # Sessions are their own entry result
        assert entry_result is session
        # Check connected session behaves as expected
        check_connected_sync_session(session)


# Sessions support implicit creation of the underlying websocket


@pytest.mark.lmstudio
def test_implicit_connection_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    api_host = Client.find_default_local_api_host()
    client = Client(api_host)
    session = _SyncSessionSystem(client)
    # Sessions start out disconnected
    assert not session.connected
    try:
        # Sync sessions will connect implicitly
        models = session.remote_call("listDownloadedModels")
        assert models is not None
        # Check connected session behaves as expected
        check_connected_sync_session(session)
    finally:
        # Still close the session even if an assertion fails
        session.close()


@pytest.mark.lmstudio
def test_implicit_reconnection_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    api_host = Client.find_default_local_api_host()
    client = Client(api_host)
    session = _SyncSessionSystem(client)
    with session:
        assert session.connected
    # Session is disconnected after use
    assert not session.connected
    try:
        # Sync sessions will reconnect implicitly
        models = session.remote_call("listDownloadedModels")
        assert models is not None
        # Check connected session behaves as expected
        check_connected_sync_session(session)
    finally:
        # Still close the session even if an assertion fails
        session.close()


# Also test the underlying websocket helper classes directly

# From RFC 6455 via
# http://python-hyper.org/projects/wsproto/en/stable/api.html#wsproto.connection.ConnectionState
WS_STATE_OPEN = 1
WS_STATE_LOCAL_CLOSING = 3
WS_STATE_CLOSED = 4
# We only expect local websocket closure, not remote
WS_CLOSING_STATES = (WS_STATE_LOCAL_CLOSING, WS_STATE_CLOSED)


@pytest.mark.asyncio
@pytest.mark.lmstudio
async def test_websocket_cm_async(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    api_host = await AsyncClient.find_default_local_api_host()
    auth_details = AsyncClient._format_auth_message()
    tm = AsyncTaskManager(on_activation=None)
    lmsws = _AsyncLMStudioWebsocket(tm, f"http://{api_host}/system", auth_details)
    # SDK client websockets start out disconnected
    assert not lmsws.connected
    # Entering the CM opens the websocket if it isn't already open
    async with tm, lmsws as entry_result:
        assert lmsws.connected
        httpx_ws = lmsws._httpx_ws
        assert httpx_ws is not None
        assert httpx_ws.connection.state.value == WS_STATE_OPEN
        # Sessions are their own entry result
        assert entry_result is lmsws
        # Attempting explicit reconnection fails when connected
        with pytest.raises(LMStudioWebsocketError, match="already connected"):
            await lmsws.connect()
        # Reentering the CM has no effect if the websocket is open
        async with lmsws:
            assert lmsws.connected
            assert httpx_ws.connection.state.value == WS_STATE_OPEN
            assert lmsws._httpx_ws is httpx_ws
        # But the websocket is closed after the *first* CM exit
        assert not lmsws.connected
        assert httpx_ws.connection.state.value in WS_CLOSING_STATES


@pytest.fixture
def ws_thread() -> Generator[AsyncWebsocketThread, None, None]:
    ws_thread = AsyncWebsocketThread()
    ws_thread.start()
    try:
        yield ws_thread
    finally:
        ws_thread.terminate()


@pytest.mark.lmstudio
def test_websocket_cm_sync(ws_thread: AsyncWebsocketThread, caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    api_host = Client.find_default_local_api_host()
    auth_details = Client._format_auth_message()
    lmsws = SyncLMStudioWebsocket(ws_thread, f"http://{api_host}/system", auth_details)
    # SDK client websockets start out disconnected
    assert not lmsws.connected
    # Entering the CM opens the websocket if it isn't already open
    with lmsws as entry_result:
        assert lmsws.connected
        httpx_ws = lmsws._httpx_ws
        assert httpx_ws is not None
        assert httpx_ws.connection.state.value == WS_STATE_OPEN
        # Sessions are their own entry result
        assert entry_result is lmsws
        # Attempting explicit reconnection fails when connected
        with pytest.raises(LMStudioWebsocketError, match="already connected"):
            lmsws.connect()
        # Reentering the CM has no effect if the websocket is open
        with lmsws:
            assert lmsws.connected
            assert httpx_ws.connection.state.value == WS_STATE_OPEN
            assert lmsws._httpx_ws is httpx_ws
        # But the websocket is closed after the *first* CM exit
        assert not lmsws.connected
        assert httpx_ws.connection.state.value in WS_CLOSING_STATES



"""Invoking and supporting prompt preprocessor hook implementations."""

import asyncio

from contextlib import asynccontextmanager
from dataclasses import dataclass
from traceback import format_tb
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    TypeAlias,
)
from typing_extensions import (
    # Native in 3.11+
    assert_never,
)

from anyio import create_task_group
from anyio.abc import TaskGroup

from ..._logging import new_logger
from ...schemas import DictObject, EmptyDict, ValidationError
from ...history import UserMessage, UserMessageDict
from ...json_api import (
    ChannelCommonRxEvent,
    ChannelEndpoint,
    ChannelFinishedEvent,
    ChannelRxEvent,
    SendMessageAsync,
    load_struct,
)
from ..._sdk_models import (
    PluginsRpcProcessingHandleUpdateParameter,
    ProcessingUpdate,
    ProcessingUpdateStatusCreate,
    ProcessingUpdateStatusUpdate,
    PromptPreprocessingAbortedDict,
    PromptPreprocessingCompleteDict,
    PromptPreprocessingErrorDict,
    PromptPreprocessingRequest,
    SerializedLMSExtendedErrorDict,
    StatusStepState,
    StatusStepStatus,
)
from ..config_schemas import BaseConfigSchema
from .common import (
    _AsyncSessionPlugins,
    HookController,
    ServerRequestError,
    StatusBlockController,
    TPluginConfigSchema,
    TGlobalConfigSchema,
)

# Available as lmstudio.plugin.hooks.*
__all__ = [
    "PromptPreprocessorController",
    "PromptPreprocessorHook",
    "run_prompt_preprocessor",
]


class PromptPreprocessingAbortEvent(ChannelRxEvent[str]):
    pass


class PromptPreprocessingRequestEvent(ChannelRxEvent[PromptPreprocessingRequest]):
    pass


PromptPreprocessingRxEvent: TypeAlias = (
    PromptPreprocessingAbortEvent
    | PromptPreprocessingRequestEvent
    | ChannelCommonRxEvent
)


class PromptPreprocessingEndpoint(
    ChannelEndpoint[tuple[str, str], PromptPreprocessingRxEvent, EmptyDict]
):
    """API channel endpoint to accept prompt preprocessing requests."""

    _API_ENDPOINT = "setPromptPreprocessor"
    _NOTICE_PREFIX = "Prompt preprocessing"

    def __init__(self) -> None:
        super().__init__({})

    def iter_message_events(
        self, contents: DictObject | None
    ) -> Iterable[PromptPreprocessingRxEvent]:
        match contents:
            case None:
                # Server can only terminate the link by closing the websocket
                pass
            case {"type": "abort", "taskId": str(task_id)}:
                yield PromptPreprocessingAbortEvent(task_id)
            case {"type": "preprocess"} as request_dict:
                parsed_request = PromptPreprocessingRequest._from_any_api_dict(
                    request_dict
                )
                yield PromptPreprocessingRequestEvent(parsed_request)
            case unmatched:
                self.report_unknown_message(unmatched)

    def handle_rx_event(self, event: PromptPreprocessingRxEvent) -> None:
        match event:
            case PromptPreprocessingAbortEvent(task_id):
                self._logger.debug(f"Aborting {task_id}", task_id=task_id)
            case PromptPreprocessingRequestEvent(request):
                task_id = request.task_id
                self._logger.debug(
                    "Received prompt preprocessing request", task_id=task_id
                )
            case ChannelFinishedEvent(_):
                pass
            case _:
                assert_never(event)


class PromptPreprocessorController(
    HookController[PromptPreprocessingRequest, TPluginConfigSchema, TGlobalConfigSchema]
):
    """API access for prompt preprocessor hook implementations."""

    def __init__(
        self,
        session: _AsyncSessionPlugins,
        request: PromptPreprocessingRequest,
        plugin_config_schema: type[TPluginConfigSchema],
        global_config_schema: type[TGlobalConfigSchema],
    ) -> None:
        """Initialize prompt preprocessor hook controller."""
        super().__init__(session, request, plugin_config_schema, global_config_schema)
        self.task_id = request.task_id
        self.pci = request.pci
        self.token = request.token

    async def _send_handle_update(self, update: ProcessingUpdate) -> Any:
        handle_update = PluginsRpcProcessingHandleUpdateParameter(
            pci=self.pci,
            token=self.token,
            update=update,
        )
        return await self.session.remote_call("processingHandleUpdate", handle_update)

    async def _create_status_block(
        self, block_id: str, status: StatusStepStatus, message: str
    ) -> None:
        await self._send_handle_update(
            ProcessingUpdateStatusCreate(
                id=block_id,
                state=StatusStepState(
                    status=status,
                    text=message,
                ),
            ),
        )

    async def _send_status_update(
        self, block_id: str, status: StatusStepStatus, message: str
    ) -> None:
        await self._send_handle_update(
            ProcessingUpdateStatusUpdate(
                id=block_id,
                state=StatusStepState(
                    status=status,
                    text=message,
                ),
            ),
        )

    async def notify_start(self, message: str) -> StatusBlockController:
        """Report task initiation in a new UI status block, return controller for updates."""
        status_block = StatusBlockController(
            self._create_ui_block_id(),
            self._send_status_update,
        )
        await self._create_status_block(status_block._id, "waiting", message)
        return status_block

    async def notify_done(self, message: str) -> None:
        """Report task completion in a new UI status block."""
        await self._create_status_block(self._create_ui_block_id(), "done", message)


PromptPreprocessorHook = Callable[
    [PromptPreprocessorController[Any, Any], UserMessage],
    Awaitable[UserMessage | UserMessageDict | None],
]


# TODO: Define a common "PluginHookHandler" base class
@dataclass()
class PromptPreprocessor(Generic[TPluginConfigSchema, TGlobalConfigSchema]):
    """Handle accepting prompt preprocessing requests."""

    plugin_name: str
    hook_impl: PromptPreprocessorHook
    plugin_config_schema: type[TPluginConfigSchema]
    global_config_schema: type[TGlobalConfigSchema]

    def __post_init__(self) -> None:
        self._logger = logger = new_logger(__name__)
        logger.update_context(plugin_name=self.plugin_name)
        self._abort_events: dict[str, asyncio.Event] = {}

    async def process_requests(
        self, session: _AsyncSessionPlugins, notify_ready: Callable[[], Any]
    ) -> None:
        """Create plugin channel and wait for server requests."""
        logger = self._logger
        endpoint = PromptPreprocessingEndpoint()
        # Async API expects timeouts to be handled via task groups,
        # so there's no default timeout to override when creating the channel
        async with session._create_channel(endpoint) as channel:
            notify_ready()
            logger.info("Opened channel to receive prompt preprocessing requests...")
            send_message = channel.send_message
            async with create_task_group() as tg:
                logger.debug("Waiting for prompt preprocessing requests...")
                async for contents in channel.rx_stream():
                    logger.debug(
                        f"Handling prompt preprocessing channel message: {contents}"
                    )
                    for event in endpoint.iter_message_events(contents):
                        logger.debug("Handling prompt preprocessing channel event")
                        endpoint.handle_rx_event(event)
                        match event:
                            case PromptPreprocessingAbortEvent():
                                await self._abort_hook_invocation(
                                    event.arg, send_message
                                )
                            case PromptPreprocessingRequestEvent():
                                logger.debug(
                                    "Running prompt preprocessing request hook"
                                )
                                ctl = PromptPreprocessorController(
                                    session,
                                    event.arg,
                                    self.plugin_config_schema,
                                    self.global_config_schema,
                                )
                                tg.start_soon(self._invoke_hook, ctl, send_message)
                    if endpoint.is_finished:
                        break

    async def _abort_hook_invocation(
        self, task_id: str, send_json: SendMessageAsync
    ) -> None:
        """Abort the specified hook invocation (if it is still running)."""
        abort_event = self._abort_events.get(task_id, None)
        if abort_event is not None:
            abort_event.set()
        response = PromptPreprocessingAbortedDict(
            type="aborted",
            taskId=task_id,
        )
        await send_json(response)

    async def _cancel_on_event(
        self, tg: TaskGroup, event: asyncio.Event, message: str
    ) -> None:
        await event.wait()
        self._logger.info(message)
        tg.cancel_scope.cancel()

    @asynccontextmanager
    async def _registered_hook_invocation(
        self, task_id: str
    ) -> AsyncIterator[asyncio.Event]:
        logger = self._logger
        abort_events = self._abort_events
        if task_id in abort_events:
            err_msg = f"Hook invocation already in progress for {task_id}"
            raise ServerRequestError(err_msg)
        abort_events[task_id] = abort_event = asyncio.Event()
        try:
            async with create_task_group() as tg:
                tg.start_soon(
                    self._cancel_on_event,
                    tg,
                    abort_event,
                    f"Aborting request {task_id}",
                )
                logger.info(f"Processing request {task_id}")
                yield abort_event
                tg.cancel_scope.cancel()
        finally:
            abort_events.pop(task_id, None)
        if abort_event.is_set():
            completion_message = f"Aborted request {task_id}"
        else:
            completion_message = f"Processed request {task_id}"
        logger.info(completion_message)

    async def _invoke_hook(
        self,
        ctl: PromptPreprocessorController[TPluginConfigSchema, TGlobalConfigSchema],
        send_json: SendMessageAsync,
    ) -> None:
        logger = self._logger
        task_id = ctl.task_id
        message = ctl.request.input
        error_details: SerializedLMSExtendedErrorDict | None = None
        response_dict: UserMessageDict
        expected_cls = UserMessage
        try:
            if not isinstance(message, expected_cls):
                err_msg = f"Received {type(message).__name__!r} ({expected_cls.__name__!r} expected)"
                raise ServerRequestError(err_msg)
            async with self._registered_hook_invocation(task_id) as abort_event:
                response = await self.hook_impl(ctl, message)
        except Exception as exc:
            err_msg = "Error calling prompt preprocessing hook"
            logger.error(err_msg, exc_info=True, exc=repr(exc))
            # TODO: Determine if it's worth sending the stack trace to the server
            ui_cause = f"{err_msg}\n({type(exc).__name__}: {exc})"
            error_details = SerializedLMSExtendedErrorDict(
                cause=ui_cause, stack="\n".join(format_tb(exc.__traceback__))
            )
        else:
            if abort_event.is_set():
                # Processing was aborted by the server, skip sending a response
                return
            if response is None:
                logger.debug("No changes made to preprocessed prompt")
                response_dict = message.to_dict()
            else:
                logger.debug(
                    "Validating prompt preprocessing response", response=response
                )
                if isinstance(response, dict):
                    try:
                        parsed_response = load_struct(response, expected_cls)
                    except ValidationError as exc:
                        err_msg = f"Failed to parse prompt preprocessing response as {expected_cls.__name__}\n({exc})"
                        logger.error(err_msg)
                        error_details = SerializedLMSExtendedErrorDict(cause=err_msg)
                    else:
                        response_dict = parsed_response.to_dict()
                elif isinstance(response, UserMessage):
                    response_dict = response.to_dict()
                else:
                    err_msg = f"Prompt preprocessing hook returned {type(response).__name__!r} ({expected_cls.__name__!r} expected)"
                    logger.error(err_msg)
                    error_details = SerializedLMSExtendedErrorDict(cause=err_msg)
        channel_message: DictObject
        if error_details is not None:
            error_title = f"Prompt preprocessing error in plugin {self.plugin_name!r}"
            common_error_args: SerializedLMSExtendedErrorDict = {
                "title": error_title,
                "rootTitle": error_title,
            }
            error_details.update(common_error_args)
            channel_message = PromptPreprocessingErrorDict(
                type="error",
                taskId=task_id,
                error=error_details,
            )
        else:
            channel_message = PromptPreprocessingCompleteDict(
                type="complete",
                taskId=task_id,
                processed=response_dict,
            )
        await send_json(channel_message)


async def run_prompt_preprocessor(
    plugin_name: str,
    hook_impl: PromptPreprocessorHook,
    plugin_config_schema: type[BaseConfigSchema],
    global_config_schema: type[BaseConfigSchema],
    session: _AsyncSessionPlugins,
    notify_ready: Callable[[], Any],
) -> None:
    """Accept prompt preprocessing requests."""
    prompt_preprocessor = PromptPreprocessor(
        plugin_name, hook_impl, plugin_config_schema, global_config_schema
    )
    await prompt_preprocessor.process_requests(session, notify_ready)



"""Command line interface implementation."""

import argparse
import logging
import os.path
import sys
import warnings

from typing import Sequence

from ..sdk_api import sdk_public_api

from . import _dev_runner, runner


def _parse_args(
    argv: Sequence[str] | None = None,
) -> tuple[argparse.ArgumentParser, argparse.Namespace]:
    py_name = os.path.basename(sys.executable).removesuffix(".exe")
    parser = argparse.ArgumentParser(
        prog=f"{py_name} -m {__spec__.parent}",
        description="LM Studio plugin runner for Python plugins",
    )
    parser.add_argument(
        "plugin_path", metavar="PLUGIN_PATH", help="Directory name of plugin to run"
    )
    parser.add_argument("--dev", action="store_true", help="Run in development mode")
    parser.add_argument("--debug", action="store_true", help="Enable debug logging")
    return parser, parser.parse_args(argv)


@sdk_public_api()
def main(argv: Sequence[str] | None = None) -> int:
    """Run the ``lmstudio.plugin`` CLI.

    If *args* is not given, defaults to using ``sys.argv``.
    """
    parser, args = _parse_args(argv)
    plugin_path = args.plugin_path
    if not os.path.exists(plugin_path):
        parser.print_usage()
        print(f"ERROR: Failed to find plugin folder at {plugin_path!r}")
        return 1
    log_level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=log_level)
    if sys.platform == "win32":
        # Accept Ctrl-C events even in non-default process groups
        # (allows for graceful termination when Ctrl-C is received
        # from a controlling process rather than from a console)
        # Based on https://github.com/python/cpython/blob/3.14/Lib/test/win_console_handler.py
        # and https://stackoverflow.com/questions/35772001/how-can-i-handle-a-signal-sigint-on-a-windows-os-machine/35792192#35792192
        from ctypes import c_void_p, windll, wintypes

        SetConsoleCtrlHandler = windll.kernel32.SetConsoleCtrlHandler
        SetConsoleCtrlHandler.argtypes = (c_void_p, wintypes.BOOL)
        SetConsoleCtrlHandler.restype = wintypes.BOOL
        if not SetConsoleCtrlHandler(None, 0):
            print("Failed to enable Ctrl-C events, termination may be abrupt")
    if not args.dev:
        warnings.filterwarnings(
            "ignore", ".*the plugin API is not yet stable", FutureWarning
        )
        try:
            runner.run_plugin(plugin_path, allow_local_imports=True)
        except KeyboardInterrupt:
            print("Plugin execution terminated by console interrupt", flush=True)
    else:
        # Retrieve args from API host, spawn plugin in subprocess
        try:
            _dev_runner.run_plugin(plugin_path, debug=args.debug)
        except KeyboardInterrupt:
            pass  # Subprocess handles reporting the plugin termination
    return 0
